Package: GridOnClusters
Type: Package
Title: Multivariate Joint Grid Discretization
Version: 0.3.2
Date: 2025-12-12
Depends: R (>= 3.5.0)
Authors@R: 
    c(person(given = "Jiandong",
             family = "Wang",
             role = "aut",
             email = "wangjd24@nmsu.edu"),
      person(given = "Sajal",
             family = "Kumar",
             role = "aut",
             email = "sajal49@nmsu.edu",
             comment = c(ORCID = "0000-0003-0930-1582")),
      person(given = "Joe",
             family = "Song",
             role = c("aut", "cre"),
             email = "joemsong@nmsu.edu",
             comment = c(ORCID = "0000-0002-6883-6547")))
Author: Jiandong Wang [aut],
  Sajal Kumar [aut] (ORCID: <https://orcid.org/0000-0003-0930-1582>),
  Joe Song [aut, cre] (ORCID: <https://orcid.org/0000-0002-6883-6547>)
Maintainer: Joe Song <joemsong@nmsu.edu>
Description: Discretize multivariate continuous data using a grid
 to capture the joint distribution that preserves clusters in
 original data. It can handle both labeled or unlabeled data.
 Both published methods (Wang et al 2020) <doi:10.1145/3388440.3412415>
 and new methods are included. Joint grid discretization
 can prepare data for model-free inference of association,
 function, or causality.
Imports: Rcpp, Ckmeans.1d.dp, cluster, fossil, dqrng, mclust, Rdpack,
        plotrix
Suggests: FunChisq, knitr, testthat (>= 2.1.0), rmarkdown
RdMacros: Rdpack
License: LGPL (>= 3)
Encoding: UTF-8
LinkingTo: BH, Rcpp
RoxygenNote: 7.3.3
NeedsCompilation: yes
VignetteBuilder: knitr
Packaged: 2025-12-12 13:19:38 UTC; joesong
Repository: CRAN
Date/Publication: 2025-12-12 13:40:07 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-27 09:53:27 UTC; unix
Archs: GridOnClusters.so.dSYM

Package: GridOnClusters
Type: Package
Title: Cluster-Preserving Multivariate Joint Grid Discretization
Version: 0.4.1
Date: 2026-07-18
Depends: R (>= 3.5.0)
Authors@R: 
    c(person(given = "Jiandong",
             family = "Wang",
             role = "aut",
             email = "wangjd24@nmsu.edu",
             comment = c(ORCID = "0009-0006-2040-7870")),
      person(given = "Sajal",
             family = "Kumar",
             role = "aut",
             email = "sajal49@nmsu.edu",
             comment = c(ORCID = "0000-0003-0930-1582")),
      person(given = "Joe",
             family = "Song",
             role = c("aut", "cre"),
             email = "joemsong@nmsu.edu",
             comment = c(ORCID = "0000-0002-6883-6547")))
Author: Jiandong Wang [aut] (ORCID: <https://orcid.org/0009-0006-2040-7870>),
  Sajal Kumar [aut] (ORCID: <https://orcid.org/0000-0003-0930-1582>),
  Joe Song [aut, cre] (ORCID: <https://orcid.org/0000-0002-6883-6547>)
Maintainer: Joe Song <joemsong@nmsu.edu>
Description: Discretize multivariate continuous data using a grid to capture the 
 joint distribution that preserves clusters in original data (Wang et al 2020) 
 <doi:10.1145/3388440.3412415>. It can handle both labeled or unlabeled data. 
 Joint grid discretization can prepare data for model-free inference of 
 association, function, or causality.
Imports: Rcpp, Ckmeans.1d.dp, cluster, fossil, dqrng, mclust, Rdpack,
        plotrix
Suggests: FunChisq, knitr, testthat (>= 2.1.0), rmarkdown
RdMacros: Rdpack
License: LGPL (>= 3)
Encoding: UTF-8
LinkingTo: BH, Rcpp, RcppArmadillo
NeedsCompilation: yes
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
Packaged: 2026-07-18 12:05:55 UTC; joesong
Repository: CRAN
Date/Publication: 2026-07-18 12:40:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(plot,GridOnClusters)
export(ballclust)
export(cluster)
export(discretize.jointly)
export(gen_simdata)
export(hashed_upsilon)
export(plotGOCpatterns)
import(Rcpp)
importFrom(Ckmeans.1d.dp,MultiChannel.WUC)
importFrom(Ckmeans.1d.dp,ahist)
importFrom(Rdpack,reprompt)
importFrom(cluster,pam)
importFrom(cluster,silhouette)
importFrom(dqrng,dqsample)
importFrom(fossil,adj.rand.index)
importFrom(grDevices,colorRampPalette)
importFrom(grDevices,palette)
importFrom(graphics,abline)
importFrom(graphics,legend)
importFrom(graphics,mtext)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,strheight)
importFrom(graphics,strwidth)
importFrom(mclust,Mclust)
importFrom(mclust,hc)
importFrom(mclust,mclustBIC)
importFrom(plotrix,color.legend)
importFrom(stats,dist)
importFrom(stats,kmeans)
importFrom(stats,pchisq)
importFrom(stats,rnorm)
useDynLib(GridOnClusters)

---
title: The 'GridOnClusters' R package
bibliography: inst/REFERENCES.bib
---
[![Project Status: Active – The project has reached a stable, usable state and is being actively developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/GridOnClusters)](https://cran.r-project.org/package=GridOnClusters)
[![CRAN_latest_release_date](https://www.r-pkg.org/badges/last-release/GridOnClusters)](https://cran.r-project.org/package=GridOnClusters)
[![metacran downloads](https://cranlogs.r-pkg.org/badges/GridOnClusters)](https://cran.r-project.org/package=GridOnClusters)
[![metacran downloads](https://cranlogs.r-pkg.org/badges/grand-total/GridOnClusters)](https://cran.r-project.org/package=GridOnClusters)



### Overview

The package offers a method to discretize multivariate continuous data using a grid that captures the joint distribution via preserving clusters in the original data (Wang, Kumar, and Song 2020). Joint grid discretization is applicable as a data transformation step before using other methods to infer association, function, or causality without assuming a parametric model.

### When to use the package

Most available discretization methods process one variable at a time, such as ['Ckmeans.1d.dp'](https://cran.r-project.org/package=Ckmeans.1d.dp). If discretizing each variable independently misses patterns arising from the joint distribution of multiple involved variables, one may benefit from using the joint discretization method in this package.

### To download and install the package

```{r}
install.packages("GridOnClusters")
```

### Examples

See the *Examples* vignette of the package.

### Citing the package

Wang J, Kumar S, Song M (2020). "Joint Grid
Discretization for Biological Pattern Discovery." In
_Proceedings of the 11th ACM International Conference
on Bioinformatics, Computational Biology and Health
Informatics_. Article no. 57. doi: 10.1145/3388440.3412415 (URL:
https://doi.org/10.1145/3388440.3412415).

---
title: The 'GridOnClusters' R package
bibliography: inst/REFERENCES.bib
---
[![Project Status: Active – The project has reached a stable, usable state and is being actively developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/GridOnClusters)](https://cran.r-project.org/package=GridOnClusters)
[![CRAN_latest_release_date](https://www.r-pkg.org/badges/last-release/GridOnClusters)](https://cran.r-project.org/package=GridOnClusters)
[![metacran downloads](https://cranlogs.r-pkg.org/badges/GridOnClusters)](https://cran.r-project.org/package=GridOnClusters)
[![metacran downloads](https://cranlogs.r-pkg.org/badges/grand-total/GridOnClusters)](https://cran.r-project.org/package=GridOnClusters)



### Overview

The package offers multiple methods to discretize multivariate 
continuous data using a grid that captures the joint 
distribution via preserving clusters in original data
(Wang, Kumar, and Song 2020). Joint grid discretization is 
applicable as a data transformation step before using 
other methods to infer association, function, or 
causality without assuming a parametric model.

### When to use the package

Most available discretization methods process one 
variable at a time, such as 
['Ckmeans.1d.dp'](https://cran.r-project.org/package=Ckmeans.1d.dp). 
As discretizing each variable independently may mis-represent 
patterns arising from the joint distribution of multiple variables, 
one may benefit from joint discretization. The methods can handle
both unlabeled and labeled data.

### To download and install the package

```{r}
install.packages("GridOnClusters")
```

### Examples

See the *Examples* vignette of the package.

### Citing the package

Wang J, Kumar S, Song M (2020). "Joint Grid
Discretization for Biological Pattern Discovery." In
_Proceedings of the 11th ACM International Conference
on Bioinformatics, Computational Biology and Health
Informatics_. Article no. 57. doi: 10.1145/3388440.3412415 (URL:
https://doi.org/10.1145/3388440.3412415).

## ----setup, include = FALSE---------------------------------------------------

knitr::opts_chunk$set(
   collapse = TRUE,
   comment = "#>"
)

## ----clustering nonlinear patterns by fixed numbers of clusters, out.width="40%", fig.show="hold", fig.cap="Example 1. Nonlinear curves using kmeans+silhouette and Ball+BIC clustering with a fixed number of clusters."----
require(GridOnClusters)
x = rnorm(500)
y = sin(x)+rnorm(500, sd = 0)
z = cos(x)+rnorm(500, sd = 0)
data = cbind(x, y, z)
ks = 2:20

res = discretize.jointly(data, k=ks, cluster_method = "kmeans+silhouette", 
                         grid_method = "Sort+split") 
plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "kmeans+silhouette", 
  grid_method = "DP exact likelihood") 
plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "Ball+BIC", 
  grid_method = "DP approx likelihood 1-way")
plot(res)

## ----clustering nonlinear patterns by varying numbers of clusters, out.width="40%", fig.show="hold", fig.cap="Example 2. Using a range for the number of kmeans+silhouette and Ball+BIC clusters"----
x = rnorm(100)
y = log1p(abs(x))
z = ifelse(x >= -0.5 & x <= 0.5, 0, 1) + rnorm(100, 0, 0.1)
data = cbind(x, y, z)
ks = c(2:5)

#res = discretize.jointly(data, k=ks, cluster_method = "Ball+BIC",
#                         grid_method = "Sort+split", min_level = 1)
#plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "kmeans+silhouette", 
  grid_method = "Sort+split", min_level = 1) 

plot(res)

#res = discretize.jointly(data, k=ks, cluster_method = "kmeans+silhouette", 
#                         grid_method = "DP exact likelihood", min_level = 1) 
#plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "Ball+BIC", 
  grid_method = "DP exact likelihood", min_level = 1) 
plot(res)

## ----Example 3 using PAM for clustering, out.width="40%", fig.show="hold", fig.cap="Example 3. Using the partition around medoids clustering method."----
# using a clustering method other than kmeans+silhouette
x = rnorm(100)
y = log1p(abs(x))
z = sin(x)
data = cbind(x, y, z)

# pre-cluster the data using partition around medoids (PAM)
cluster_label = cluster::pam(x=data, diss = FALSE, metric = "euclidean", k = 4)$clustering

res = discretize.jointly(
  data, cluster_label = cluster_label,
  grid_method = "Sort+split", min_level = 1)
res = discretize.jointly(
  data, cluster_label = cluster_label,
  grid_method = "DP exact likelihood", min_level = 1)
plot(res, main="Original data\nPAM clustering", 
     main.table="Discretized data\nPAM & Sort+split")

## ----Example 4, out.width="40%", fig.show="hold", fig.cap="Example 4. Random patterns using kmeans+silhouette and Ball+BIC clustering with a range."----
ks = 2:20
n = 40*10
sd = 60*4

x=rnorm(2*n, sd=sd)
y=rnorm(2*n, sd=sd)
x=c(x,rnorm(2*n, sd=sd/3))
y=c(y,rnorm(2*n, sd=sd/3)+200)

data = cbind(x, y)

res = discretize.jointly(
  data, k=ks, cluster_method = "Ball+BIC",
  grid_method = "Sort+split", min_level = 1)
plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "kmeans+silhouette", 
  grid_method = "Sort+split", min_level = 1) 
plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "kmeans+silhouette", 
  grid_method = "DP approx likelihood 1-way", min_level = 1) 
plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "Ball+BIC",
  grid_method = "DP approx likelihood 1-way", min_level = 1)
plot(res)

## ----Example 5 bivariate, out.width="40%", fig.show="hold", fig.cap="Example 5. Multi-cluster random patterns using kmeans+silhouette and Ball+BIC clustering with a range."----
n <- 50*8
ks <- 2:20

X.C1 <- matrix(
   c(rnorm(n, 5, sd=2),
     rnorm(n, 0, sd=40)), 
   ncol = 2, byrow = FALSE
)
X.C2 <- matrix(
   c(rnorm(n, 70, sd=1),
     rnorm(n, 0, sd=1)), 
   ncol = 2, byrow = FALSE
)

X.C3 <- matrix(
   c(rnorm(n, 150, sd=30),
     rnorm(n, 0, sd=30)), 
   ncol = 2, byrow = FALSE
)

data = rbind(X.C1, X.C3)

res = discretize.jointly(
  data, k=ks, cluster_method = "Ball+BIC",
  grid_method = "Sort+split", min_level = 1)
plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "kmeans+silhouette", 
  grid_method = "Sort+split", min_level = 1) 
plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "kmeans+silhouette", 
  grid_method = "DP approx likelihood 1-way", 
  min_level = 1, cutoff = 1) 
plot(res)
res = discretize.jointly(
  data, k=ks, cluster_method = "Ball+BIC", 
  grid_method = "DP approx likelihood 1-way", 
  min_level = 1) 
plot(res)

## ----Example 6. Exclusive or, out.width="40%", fig.show="hold", fig.cap="Example 6. Exclusive or. Dim1 ⊕ Dim2, Dim3 and Dim4 are random"----
n <- 100
ks <- 2:10

label = c(rep(1, n), rep(2,n), rep(3,n), rep(4,n))

X1 = c(rnorm(n, 0, sd=2),
     rnorm(n, 0, sd=2),
     rnorm(n, 10, sd=2),
     rnorm(n, 10, sd=2))

X2 = c(rnorm(n, 10, sd=2),
     rnorm(n, 0, sd=2),
     rnorm(n, 10, sd=2),
     rnorm(n, 0, sd=2))

X3 = c(rnorm(4*n, 20, sd=10))

X4 = c(rnorm(4*n, 3, sd=20))

data = cbind(X1, X2, X3, X4)

#res = discretize.jointly(data, k=ks, cluster_method = "kmeans+silhouette", 
#                         grid_method = "DP approx likelihood", min_level = 1) 
#plot(res)
#
#res = discretize.jointly(data, k=ks, cluster_method = "kmeans+silhouette", 
#                         grid_method = "DP Compressed majority", min_level = 1) 
#plot(res)

#res = discretize.jointly(data, k=ks, cluster_method = "Ball+BIC", 
#                         grid_method = "DP exact likelihood", min_level = 1) 
#plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "Ball+BIC", 
  grid_method = "DP approx likelihood 1-way", 
  min_level = 1) 
plot(res)

res = discretize.jointly(
  data, k=ks, cluster_label = label,
  grid_method = "DP approx likelihood 1-way", 
  min_level = 1) 
plot(res)

#res = discretize.jointly(data, k=ks, cluster_label = label,
#                         grid_method = "Sort+split", min_level = 1) 
#plot(res)
#
#res = discretize.jointly(data, k=ks, cluster_method = "kmeans+silhouette", 
#                         grid_method = "Sort+split", min_level = 1) 
#plot(res)


## ----Example 7. Three well separated rounds., out.width="40%", fig.show="hold", fig.cap="Example 7. Three rounds well seperated on x axis"----
n <- 20
ks <- 2:10

label = c(rep(1, n), rep(2,5*n), rep(3, 5*n))

X1 = c(rnorm(n, 0, sd=1), rnorm(5*n, 15, sd=3), rnorm(5*n, 35, sd=3))
X2 = c(rnorm(n, 0, sd=1), rnorm(5*n, 0, sd=3), rnorm(5*n, 0, sd=3))

data = cbind(X1, X2)

res = discretize.jointly(
  data, cluster_label = label,
  grid_method = "DP exact likelihood", min_level = 1) 
plot(res)
res = discretize.jointly(
  data,  cluster_label = label,
  grid_method = "DP approx likelihood 1-way", 
  min_level = 1) 
plot(res)

res.entropy = discretize.jointly(
  data,  cluster_label = label,
  grid_method = "DP exact likelihood", min_level = 1, entropy = TRUE) 
plot(res.entropy)



## ----Example 8. Four spheres with varying centers and radius, out.width="40%", fig.show="hold", fig.cap="Example 8. Four sphers with different centers and radius"----
n <- 200
ks <- 2:10

label = c(rep(1, n), rep(2,5*n), rep(3,n), rep(4,3*n))

X1 = c(rnorm(n, 0, sd=3), rnorm(5*n, 15, sd=3), rnorm(n, 5, sd=2), rnorm(3*n, 15, sd=4))

X2 = c(rnorm(n, 0, sd=4), rnorm(5*n, 15, sd=3), rnorm(n, 10, sd=2), rnorm(3*n, 0, sd=4))

data = cbind(X1, X2)

res = discretize.jointly(
  data, k=ks, cluster_method = "Ball+BIC",
  grid_method = "DP exact likelihood", min_level = 1) 
plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "Ball+BIC",
  grid_method = "DP approx likelihood 1-way", 
  min_level = 1) 
plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "Ball+BIC",
  grid_method = "DP exact likelihood", min_level = 1, entropy = TRUE) 
plot(res)



## ----Example 9. Two overlapping spheres, out.width="40%", fig.show="hold", fig.cap="Example 9. A small dense sphere overlapping a large sphere."----
n <- 200
ks <- 2:10

label = c(rep(1, n), rep(2,5*n))

X1 = c(rnorm(n, 0, sd=1), rnorm(5*n, 6, sd=3))
X2 = c(rnorm(n, 0, sd=1), rnorm(5*n, 0, sd=3))

data = cbind(X1, X2)

res = discretize.jointly(
  data, k=ks, cluster_label = label,
  grid_method = "DP exact likelihood", min_level = 1) 
plot(res)

res = discretize.jointly(
  data, k=ks,  cluster_label = label,
  grid_method = "DP approx likelihood 1-way", 
  min_level = 1) 
plot(res)

res.entropy = discretize.jointly(
  data, k=ks,  cluster_label = label,
  grid_method = "DP exact likelihood", min_level = 1, entropy = TRUE) 
plot(res.entropy)

# GridOnClusters/tools.R
#
# Created  by Jiandong Wang
# Copyright (c) NMSU Song lab

#' @importFrom stats pchisq rnorm
#' @importFrom cluster pam
#' @importFrom mclust hc
#' 
#' @title Generate Simulated Data
#' @param cord data matrix that records the index for
#'   each cluster on each dimension
#' @param sim_table a matrix
#' @param noise a numeric value to specify noise level
#' @param plot a logical to turn on or off plotting
#' @export
gen_simdata = function(cord, sim_table, noise=0.3, plot = FALSE){
  
   # n is the number of points of the sample
   n <- sum(sim_table)
   
   rsums = rowSums(sim_table)
   csums = colSums(sim_table)
   ndim = ncol(cord)
   # ndim does not include the Y axis, so total dimension will be ndim+1
   dist = 5- 5*noise
   data = matrix(nrow = n, ncol = ndim+1)
   
   label = c()
   cur_l = 1
   y = c()
   x = c()
   
   start_indices <- cumsum(c(1, rsums[-length(rsums)])) 
   end_indices <- cumsum(rsums)
   if(end_indices[nrow(sim_table)] != n){
      print("Error: end_indices does not match data size(n)!")
      return(-1)
   }
   
   if(ndim>=2){
      x_dim = c()
      for (i in 1:nrow(sim_table)) {
         x_temp = c()
         y_temp = c()
         
         for (j in 1:ndim) {
            dim.index = cord[i,j]
            if(start_indices[i] <= end_indices[i]){
               data[start_indices[i]:end_indices[i], j] = rnorm(rsums[i], dim.index*dist)
            }
         }
         
         for(j in 1:ncol(sim_table)){
            y_temp = c(y_temp, rnorm(sim_table[i,j], j*dist))
            label = c(label, rep(cur_l, sim_table[i,j]))
            cur_l = cur_l+1
         }
         y = c(y, y_temp)
      }
      data[,ndim+1] = y
   }
   else{
      for (i in 1:nrow(sim_table)) {
         x_sub = rnorm(rsums[i], dist*i)
         x = c(x, x_sub)
         for (j in 1:ncol(sim_table)) {
            label = c(label, rep(cur_l,sim_table[i,j]))
            y_sub = rnorm(sim_table[i,j],dist*j)
            y = c(y, y_sub)
            cur_l = cur_l+1
         }
      }
      # why max(y)+3-y? what this works?
      # y = max(y)+3-y
      if(plot){
         plot(x,y)
      }
      data = cbind(x,y)
   }
   
   res = structure(list(sim_data=data, label = as.numeric(as.factor(label))))
   return(res)
}

#' Hashed Upsilon Test of Association
#' 
#' @param points a matrix of observed categorical data points. 
#'   Rows are observations and columns are variables.
#' @param log.p a logical. If \code{TRUE}, the returned p.value is 
#'   logarithm of the true p-value. Useful for improved precision
#'   when true p-values are close to zero.
#' @seealso \link[Upsilon]{upsilon.test} for full description of 
#'   the Upsilon test.
#' @export
hashed_upsilon = function(points, log.p=FALSE){
   if((is.matrix(points) || is.data.frame(points)) &&
      nrow(points) > 1 && ncol(points) > 1) {
      res <- upsilon_c(as.matrix(points), log.p)
   } else {
      res <- list(0, 0, 1, 0)
   }
   
   max_upsilon <- (prod(apply(points, 2, function(col) length(unique(col))))*(nrow(points)))/(2^(ncol(points)))
   urelt <- res[[1]] / max_upsilon
   prelt <- pchisq(urelt, df = 1, lower.tail = FALSE, log.p = TRUE)
   res$stat_relt <- urelt
   res$p.value_relt <- prelt
   res$tsp <- prod(apply(points, 2, function(col) length(unique(col))))
   res$tsm <- mean(apply(points, 2, function(col) length(unique(col))))
   
   full_max <- 2 * sqrt(max_upsilon) -1
   urelt <- (res[[1]] / nrow(points)) / full_max
   
   names(res) <- c("statistic", "parameter", "p.value", "estimate", "stat_relt", "p.value_relt", "tsp", "tsm")
   names(res$statistic) <- "Upsilon"
   names(res$parameter) <- "df"
   names(res$estimate) <- "Effect size"
   names(res$p.value) <- ifelse(log.p, "log P", "P")
   names(res$stat_relt) <- "Upsilon relative"
   names(res$p.value_relt) <- ifelse(log.p, "log P relative", "P relative")
   names(res$tsp) <- "table_size_production"
   names(res$tsm) <- "table_size_mean"

   level <- res[[1]] / max_upsilon
   res$upsilon_level <- level
   return(res)
}

#' NO @export
#' @keywords internal
#' @noRd
.median_distance = function(data, label){
   dist.temp <- list()
   dist.temp$distance_to_median <- median_distance_c(data, label)
   dist.temp$mean_dist_to_median <- mean(unlist(dist.temp$distance))
   return(dist.temp)
}

#' NO @export
#' @keywords internal
#' @noRd
.mean_distance = function(data, label){
   dist.temp <- list()
   dist.temp$distance_to_mean <- mean_distance_c(data, label)
   dist.temp$mean_dist_to_mean <- mean(unlist(dist.temp$distance))
   return(dist.temp)
}

# NO @export
fill_columns <- function(mat) {
   existing_cols <- as.numeric(colnames(mat))
   full_range <- seq(min(existing_cols), max(existing_cols))  # Get full range of column indices
   missing_cols <- setdiff(full_range, existing_cols)  # Find missing columns
    
   if (length(missing_cols) > 0) {
      zero_mat <- matrix(0, nrow = nrow(mat), ncol = length(missing_cols))
      colnames(zero_mat) <- as.character(missing_cols)
      
      mat <- cbind(mat, zero_mat)
      mat <- mat[, order(as.numeric(colnames(mat)))]
   }
   return(mat)
}

#' Fast VII Gaussian Mixture Model
#'
#' @param data A numeric matrix or data frame.
#' @param G An integer vector specifying the numbers of mixture components. 
#'        If NULL, defaults to 1:9.
#' @param threads Number of threads for OpenMP parallelization.
#' @param min_pts an integer to specify the minimum number of points inside 
#'        a cluster. Default: 5
#' @param verbose logical. if \code{TRUE}, print out additional messages. 
#'        Default: \code{FALSE} 
#' @return An object of class \code{ballclust}, mimicking \code{mclust}.
#' @export
ballclust <- function(data, G = NULL, threads = parallel::detectCores() - 1, 
                      min_pts = 5, verbose = FALSE) {
   X <- as.matrix(data)
   
   # Auto K-selection fallback
   if (is.null(G)) {
      G <- 1:9
   }
   
   # Call the Rcpp backend
   res <- fast_VII_mclust(X, G, threads = threads, min_pts = min_pts, verbose = verbose)
   
   # --- ADD THIS SAFETY CHECK ---
   if (is.null(res$z)) {
      stop(paste("Clustering failed for all requested K values (", paste(G, collapse=","), ").",
                 "This usually occurs because k-means failed to initialize or 
                 clusters collapsed below the 'min_pts' ", min_pts, " threshold."))
   }
   
   # Determine hard classifications
   classification <- max.col(res$z)
   
   # Format output to match mclust structure
   out <- list(
      call = match.call(),
      modelName = "VII",
      n = nrow(X),
      d = ncol(X),
      G = res$K,
      BIC = res$BIC,
      loglik = res$loglik,
      df = (res$K - 1) + (res$K * ncol(X)) + res$K,
      bic = res$BIC_all, # All tested BICs
      parameters = list(
         pro = as.numeric(res$pro),
         mean = t(res$mean), # mclust expects p x K
         variance = list(
            modelName = "VII",
            d = ncol(X),
            G = res$K,
            sigmasq = as.numeric(res$variance), # Vector of length K
            scale = as.numeric(res$variance)
         )
      ),
      z = res$z,
      classification = classification
   )
   
   class(out) <- c("mclustVII", "mclust")
   return(out)
}

# plot_patterns.R
# Created by Mingzhou Song
# Modified by Jiandong Wang and Sajal Kumar
#    Dec 6, 2021. Added function plot.GridOnClusters()
# Copyright (c) NMSU Song lab

#' Plotting the continuous data along with cluster-preserving Grid
#' 
#' Plots examples of jointly discretizing continuous data based on grids that 
#' preserve clusters in the original data.
#' 
#' @importFrom grDevices colorRampPalette palette
#' @importFrom graphics abline legend mtext par plot strheight strwidth
#' @importFrom plotrix color.legend
#' 
#'
#' @param x the result generated by discretize.jointly
#' @param xlab the horizontal axis label
#' @param ylab the vertical axis label
#' @param main the title of the clustering scatter plots
#' @param main.table the title of the discretized data plots
#' @param sub the subtitle
#' @param pch the symbol for points on the scatter plots
#' @param col the color of data points
#' @param line_col the color of grid lines
#' @param cex A numerical value giving the amount by 
#' which plotting text and symbols should be magnified 
#' relative to the default.
#' @param plot.table a logical to show the contingency 
#'   table. Default: \code{TRUE}.
#' @param ... additional graphical parameters
#' 
#' @export
plot.GridOnClusters = function(
   x, xlab=NULL, ylab=NULL, main=NULL, 
   main.table=NULL, col, line_col="black",
   cex = 1.125, sub=NULL, pch = 19, 
   plot.table = TRUE, ...)
{
   k = length(unique(x$clabels))
   mar = c(2.5,2.5,2.5,3.5)
   par(mar=mar, mgp=c(3,1,0)-c(1.5,0.5,0), lwd=2)
   dims = dim(x$data)[2]

   #col.palette = palette()[-1]
   l = 0
   for(i in c(1:(dims-1))){
      for(j in c((i+1):dims)){
         l = l + 1
         
         #if(is.null(xlab)) xlab <- paste0("dimension ",i)
         #if(is.null(ylab)) ylab <- paste0("dimension ",j)
         if(is.null(colnames(x$data))){
            xlab <- paste0("dimension ",i)
            ylab <- paste0("dimension ",j)
         } 
         else{
            xlab <- colnames(x$data)[i]
            ylab <- colnames(x$data)[j]
         }
         if(is.null(main)) main <- paste0("Original Data\nMethod = ", x$cluster_method, " & ", x$grid_method)
         if(is.null(main.table)) main.table <- paste0("Discretized Data\nMethod = ", x$cluster_method, " & ", x$grid_method)
               
         #col <- col.palette[l]
         #labelcol <- colorRampPalette(c("black", col))
         #plot(x$data[,i], x$data[,j], main="", col=labelcol(k)[x$clabels], 
         #     pch=19, cex.axis=0.8, cex=0.7, xlab = xlab, ylab = ylab)
         if (missing(col)){
            col=x$clabels
         }
         plot(x$data[,i], x$data[,j], main="", col=col, 
              pch=pch, cex.axis=0.8, cex=0.7, xlab = xlab, ylab = ylab)
         mtext(text = main, side = 3, line = 0.25, cex=cex)
         abline(v=x$grid[[i]], h=x$grid[[j]], col=line_col, lty="dotted")
         tab <- as.matrix(table(-x$D[, j], x$D[, i]))
         coord <- par("usr")
         #color.legend(xl = coord[2]+strwidth("0")*3,
         #             xr = coord[2]+strwidth("0")*5,
         #             yb = mean(c(coord[4],coord[3])), 
         #             yt = coord[4], 
         #             legend = seq(length(unique(x$clabels))),
         #             rect.col = unique(labelcol(k)[seq(length(unique(x$clabels)))]), 
         #             gradient="y")
         if(plot.table){
            FunChisq::plot_table(
               tab, xlab=xlab, ylab=ylab, # col=x$clabels, 
               main=main.table, highlight="none", value.cex=4/max(dim(tab)),
               mar = c(2.5, 2.5, 2.5, 2.5))
         }
      }
   }
}


#' (OBOSOLETE) Plotting the continuous data along with cluster-preserving Grid
#' 
#' Plots examples of jointly discretizing continuous data based on grids that 
#' preserve clusters in the original data.
#' 
#' @param data the input continuous data matrix
#' @param res the result generated by discretize.jointly
#' 
#' @export
plotGOCpatterns = function(data, res)
{
   k = length(unique(res$clabels))
   mar = c(2.5,2.5,2.5,3.5)
   par(mar=mar, mgp=c(3,1,0)-c(1.5,0.5,0), lwd=2)
   dims = dim(data)[2]
   
   col.palette = palette()[-1]
   l = 0
   for(i in c(1:(dims-1))){
      for(j in c((i+1):dims)){
         l = l + 1
         col <- col.palette[l]
         labelcol <- colorRampPalette(c("black", col))
         plot(data[,i], data[,j], main="", col=labelcol(k)[res$clabels], 
              pch=19, cex.axis=0.8, cex=0.7, xlab = paste0("dimension ",i), ylab = paste0("dimension ",j))
         mtext(text = "Original Data", side = 3, line = 1, cex=1.5)
         abline(v=res$grid[[i]], h=res$grid[[j]], col="black", lty="dotted")
         tab <- as.matrix(table(-res$D[, j], res$D[, i]))
         coord <- par("usr")
         color.legend(xl = coord[2]+strwidth("0")*3,
                      xr = coord[2]+strwidth("0")*5,
                      yb = mean(c(coord[4],coord[3])), 
                      yt = coord[4], 
                      legend = seq(length(unique(res$clabels))),
                      rect.col = unique(labelcol(k)[seq(length(unique(res$clabels)))]), 
                      gradient="y")
         FunChisq::plot_table(
            tab, xlab=paste0("dimension ",i), ylab=paste0("dimension ",j),
            col=col, main="Discretized data", highlight="none", 
            mar = c(2.5, 2.5, 2.5, 2.5))
      }
   }
}

# discretize.jointly.R
#
# Created by Sajal Kumar
# Modified by Jiandong Wang
# Copyright (c) NMSU Song lab

#' Discretize Multivariate Continuous Data by a Cluster-Preserving Grid
#'
#' Discretize multivariate continuous data using a grid that captures the joint distribution via
#' preserving clusters in the original data
#'
#' @importFrom cluster silhouette
#' @importFrom stats kmeans
#' @importFrom fossil adj.rand.index
#' @importFrom stats dist
#' @importFrom dqrng dqsample
#' @importFrom Rdpack reprompt
#' @importFrom mclust Mclust mclustBIC
#' @importFrom Ckmeans.1d.dp MultiChannel.WUC ahist
#@importFrom Telescope telescope
#' @import Rcpp
#' @useDynLib GridOnClusters
#'
#' @param data a matrix containing two or more continuous variables.
#' Columns are variables, rows are observations.
#'
#' @param k either an integer, a vector of integers, or \code{Inf}, specifying 
#' different ways to find clusters in \code{data}. The default is a vector 
#' containing integers from 2 to 10. If 'k' is a single number, \code{data} will 
#' be grouped into into exactly 'k' clusters. If 'k' is an integer vector, an 
#' optimal 'k' is chosen from among the integers, that maximizes the average 
#' silhouette width. If 'k' is set to \code{Inf}, an optimal k is chosen among
#' 2 to \code{nrow(data)}. If \code{cluster_label} is specified, \code{k} is 
#' ignored.
#'
#' @param cluster_label a vector of user-specified cluster labels for each observation
#' in \code{data}. The user is free to choose any clustering.
#' If unspecified, k-means clustering is used by default.
#'
#' @param min_level integer or vector, signifying the minimum number of levels 
#' along each dimension. If a vector of size \code{ncol(data)}, then each element
#' will be mapped 1:1 to each dimension in order. If an integer, then all dimensions
#' will have the same minimum number of levels.
#'
#' @param max_level integer or vector, signifying the maximum number of levels 
#' along each dimension. Works in the same way as \code{min_level}. Max_level will
#' be set to the smaller one between the number of compressed zones and itself, if 
#' the grid_method is set to "dp compressed likelihood" or "dp compressed majority"
#'
#' @param grid_method a character string to specify the discretization method. 
#' Default: \code{"DP approx likelihood 1-way"}. The methods 
#' can be roughly separate into three different categories: 
#' by cluster likelihood, by density, and by SSE (Sum of Squared Errors).
#' 
#' See Details for more information.
#'  
#' @param noise adding jitter noise to the original data, only will be used when
#' the \code{cluster_method} set to "Ball+BIC". When the data are very dense, 
#' adding noise could improve the Mclust's result.  
#'  
#' @param cluster_method the clustering method to be used. Ignored if cluster 
#' labels are given.
#' 
#' \code{"kmeans+silhouette"} will use k-means to cluster \code{data} and the average 
#'       Silhouette score to select the number of clusters k.
#' 
#'\code{"Ball+BIC"}: Uses a custom, multithreaded spherical clustering method. 
#'       It builds isotropic, variable-radius clusters and uses the BIC score to 
#'       automatically select the optimal k.
#' 
#' \code{"mclust_VII"}: Uses the \code{mclust} package (\code{modelNames = "VII"}) 
#'       to perform spherical clustering, also using BIC for k selection.
#' 
#' \code{"PAM"} pam is partition around medoids   
#' 
#' See Details for more information.
#' 
#' @param cutoff a floating number. Tell the quality to decide whether cut that 
#' line or not. If the statistic, a value to evaluate the quality of a cut,is 
#' smaller then cut_off, no cut will be performed. Only will be used when the 
#' \code{grid_method} is set to "DP" or "DP Compressed majority"
#' 
#' @param entropy Boolean variable decide whether using entropy or likelihood, 
#' default value is false.
#' 
#' @param eval_method a character string to 
#'   specify the method to evaluate quality 
#'   of discretized data
#'   
#' @param dim_reduction a logical to turn on/off 
#'   dimension reduction. Default: \code{FALSE}.
#'   
#' @param scale a logical to specify linear 
#'   scaling of the variable in each dimension.
#'   Default: \code{FALSE}.
#'   
#' @param variance a numeric value to specify 
#'   noise variance to be added to the data

#' @param nthread Integer. Number of CPU threads to use. 
#' Automatically adjusted if invalid or 
#' exceeds available cores. Default: 1
#'
#' @param min_pts an integer to specify the minimum number of points inside 
#'        a cluster. Default: 5
#'
#' @param debug logical. If \code{TRUE}, it enables the debug mode.
#'  Default: \code{FALSE}
#' @note
#' The default value of \code{grid_method} is changed
#' from "Sort+Split" (up to released package version 0.1.0.2) 
#' to "DP approx likelihood 1-way" representing a much 
#' improved approach.
#' 
#' For the default configuration, it will also add some jitter noise to the data randomly only for clustering
#' to avoid clustering failure by too many overlapping points in the data.
#' 
#' For more details, please check the documents.
#' 
#' 
#' @details 
#' 
#' The function implements both published algorithms described in 
#' \insertCite{Jwang2020BCB}{GridOnClusters} and new algorithms for
#' multivariate discretization.
#' 
#' The grid methods provided is roughly in three different kinds. 
#' 
#' * By Density
#'
#'     - \code{"Sort+split"} \insertCite{Jwang2020BCB}{GridOnClusters} 
#'     will sort the cluster by cluster mean in each dimension and then
#'  split consecutive pairs only if the sum of the error rate of each cluster is
#'  less than or equal to 50%. It is possible that no grid line will be added
#'  in a certain dimension. The maximum number of lines is the number of 
#'  clusters minus one.
#'  
#' * By SSE (Sum of Squared Errors)
#'     - \code{"MultiChannel.WUC"} will split each dimension by weighted with-in cluster
#'  sum of squared distances by "Ckmeans.1d.dp::MultiChannel.WUC". Applied in 
#'  each projection on each dimension. The channel of each point is defined by 
#'  its multivariate cluster label.
#'  
#'     - \code{"DP"} will order the label by the data in each dimension and then cut the data
#'  into max of max_level bins, then evaluating the quality of each cut to find 
#'  the best number of bins.
#'  
#'     - \code{"DP Compressed majority"} will order the label by the data in each dimension 
#'  and then compress the labels that neighbored by the same label to avoid trying 
#'  discretization within consecutive points of the same cluster label to greatly 
#'  reduce runtime of dynamic programming. Then cut the data into max of max_level 
#'  bins, and evaluating the quality of each cut by the majority of the data has 
#'  been spliced clearly to find the best number of bins.
#'  
#' * By cluster likelihood
#'     - \code{"DP exact likelihood"} will order the label by the data in each dimension
#'  and then compress the labels that neighbored by the same label to avoid trying 
#'  discretization within consecutive points of the same cluster label to greatly 
#'  reduce runtime of dynamic programming. Then cut the data into max of max_level 
#'  bins.
#'  
#'     - \code{"DP approx likelihood 1-way"} is a speed up version of the 
#'  \code{"DP exact likelihood"} method, but no optimal will be guaranteed.
#' 
#'     - \code{"DP approx likelihood 2-way"} is a bidirectional variant of the 
#'  \code{"DP approx likelihood"} method. It performs approximate dynamic 
#'  programming cuts in both the forward and backward directions and selects 
#'  the better of the two results. This approach provides additional robustness 
#'  compared to the one-directional version, but optimality is still not guaranteed.
#'  
#' The clsuter methods provided:
#' 
#'     - \code{"Ball+BIC"}: Employs a custom, high-performance spherical clustering 
#' algorithm. It models clusters as isotropic spheres (equal shape, variable volume) 
#' and automatically selects the optimal number of clusters (k) by maximizing 
#' the Bayesian Information Criterion (BIC). This implementation supports 
#' multithreading for scalable execution on high-dimensional data.
#' 
#'     - \code{"mclust_VII"}: Utilizes the \code{mclust} package to fit a Gaussian 
#' Mixture Model with a spherical, unequal volume covariance structure 
#' (\code{modelNames = "VII"}). It relies on the BIC score to automatically 
#' determine the optimal number of components.
#' 
#' @return
#'
#' A list that contains four items:
#' \item{\code{D}}{a matrix that contains the discretized version of the original \code{data}.
#' Discretized values are one(1)-based.}
#'
#' \item{\code{grid}}{a list of vectors containing decision boundaries for each variable/dimension.}
#'
#' \item{\code{clabels}}{a vector containing cluster labels for each observation in \code{data}.}
#'
#' \item{\code{csimilarity}}{a similarity score between clusters from joint discretization
#' \code{D} and cluster labels \code{clabels}. The score is the adjusted Rand index.}
#' 
#' @examples
#' # using a specified k
#' x = rnorm(100)
#' y = sin(x)
#' z = cos(x)
#' data = cbind(x, y, z)
#' discretized_data = discretize.jointly(data, k=5)$D
#'
#' # using a range of k
#' x = rnorm(100)
#' y = log1p(abs(x))
#' z = x^2 # tan(x)
#' data = cbind(x, y, z)
#' discretized_data = discretize.jointly(data, k=c(3:10))$D
#' 
#' # using k = Inf
#' x = c()
#' y = c()
#' mns = seq(0,1200,100)
#' for(i in 1:12){
#'   x = c(x,runif(n=20, min=mns[i], max=mns[i]+20))
#'   y = c(y,runif(n=20, min=mns[i], max=mns[i]+20))
#' }
#' data = cbind(x, y)
#' discretized_data = discretize.jointly(data, k=Inf)$D
#'
#' # using an alternate clustering method to k-means
#' library(cluster)
#' x = rnorm(100)
#' y = log1p(abs(x))
#' z = sin(x)
#' data = cbind(x, y, z)
#'
#' # pre-cluster the data using partition around medoids (PAM)
#' cluster_label = pam(x=data, diss = FALSE, metric = "euclidean", k = 5)$clustering
#' discretized_data = discretize.jointly(data, cluster_label = cluster_label)$D
#'
#' #
#' @references
#' \insertAllCited{}
#' 
#' @author 
#' 
#' Jiandong Wang, Sajal Kumar and Mingzhou Song
#' 
#' @seealso
#'
#' See \link[Ckmeans.1d.dp]{Ckmeans.1d.dp} for discretizing univariate continuous data.
#'
#' @md
#'
#' @export
discretize.jointly = function(
    data, k=c(2:10), min_level = 1, max_level= 100,
    cluster_method = c(
      "Ball+BIC",
      "mclust_VII", 
      "kmeans+silhouette", 
      "PAM" # , 
      # "telescope",
      # "Ball+Tele+BIC"
    ),
    grid_method = c(
      "DP approx likelihood 1-way", 
      "DP approx likelihood 2-way",
      "DP exact likelihood", 
      "DP Compressed majority", 
      "DP", 
      "Sort+split", 
      "MultiChannel.WUC"),
    eval_method = c(
      "ARI", "purity", "upsllion", "CAIR"
    ),
    cluster_label=NULL, cutoff=0.00, 
    entropy=FALSE,
    noise = FALSE, 
    dim_reduction = FALSE, 
    scale=FALSE,
    variance = 0.5, 
    min_pts = 5,
    nthread=1,
    debug=FALSE){

  # check if data provided is a matrix
  if( !("matrix" %in% class(data)) && !("data.frame" %in% class(data))){
    stop("'data' must be a matrix or data.frame.")
  }

  # check if all columns in 'data' are numeric or integer
  dim_class = apply(data, 2, class)
  if(!all(dim_class == "numeric" | dim_class == "integer")){
    stop("All columns in 'data' should be numeric or integer.")
  }

  # 'data' should have at least 10 observations
  dim_data = dim(data)
  #if(dim_data[1] < 10){
  #  stop("'data' should have at least 10 observations,")
  #}

  # 'cluster_label' should either be null or integers (or numeric) matching nrow(data)
  if(!is.null(cluster_label) && length(cluster_label) != nrow(data)){
    stop("'cluster_label' should either be null or a vector with nrow(data) elements.")
  }

  if(!is.null(cluster_label) && !class(cluster_label) %in% c("numeric","integer")){
    stop("'cluster_label' should be either null or a numeric/integer vector.")
  }
  
  # if "min_level" is a list or a integer
  if(length(min_level)==1){
    min_level = rep(min_level, ncol(data))
  } else if(length(min_level) > 1 && length(min_level) != ncol(data)){
    stop("'min_level' should either be an integer or a vector of size ncol(data)")
  }
  
  # if "max_level" is a list or a integer
  if(length(max_level)==1){
     max_level = rep(max_level, ncol(data))
  } else if(length(max_level) > 1 && length(max_level) != ncol(data)){
     stop("'min_level' should either be an integer or a vector of size ncol(data)")
  }
    
  # switching order, checking min_level < k after having a k
  
  # 'min_level' should smaller then the max of k
  if(max(k) < max(min_level)){
    k = c(min_level,min_level+5)
    warning("'min_level' should be in the range of k, k has been adapted")
   } 
  
  # the maximum of k range should be less then the number of point
  if(all(is.finite(k))){
    k = k[which(k<nrow(data))]  
  }
  
  #scale data
  data_scaled = data
  if(scale == TRUE){
     data_scaled = scale(data)
  }
  #check method
  cluster_method = match.arg(cluster_method)
  grid_method = match.arg(grid_method)
  eval_method = match.arg(eval_method)
  
  # if no cluster labels and methods are supplied, default to mclust(VII)
  if(is.null(cluster_label)){

    cluster_info =  cluster(data_scaled, k, cluster_method, noise, min_pts, nthread)

  } else { # cluster labels are supplied

    cluster_label = as.numeric(as.factor(cluster_label)) # making labels consecutive

    # # find medoids
    # centers = data[medoids(D = dist(data), cl = cluster_label),,drop=FALSE]

    cluster_info = list(#centers = as.matrix(centers),
                        clusters = cluster_label-1,
                        data = as.matrix(data), cluster_method = "user supply",
                        add_noise = FALSE)
  }

  #browser()
  if(grid_method == "MultiChannel.WUC"){  
    # prepare the weights
    # kstar = length(unique(cluster_info$clusters)) # MS 2022-01-17
    clusters.factor <- factor(cluster_info$clusters) # MS 2022-01-17
    kstar <- nlevels(clusters.factor)                # MS 2022-01-17
    weight = matrix(1/nrow(data), nrow = nrow(data), ncol = kstar)
    for (i in c(1:nrow(data))) {
      # weight[i, cluster_info$clusters[i]+1] = nrow(data)/(nrow(data)+1) # MS 2022-01-17
       weight[i, clusters.factor[i]] = nrow(data)/(nrow(data)+1) # MS 2022-01-17
    }
    grid_lines = vector("list", length = ncol(data))
    # for n dim
    for(dim in c(1:ncol(data))){
      result = Ckmeans.1d.dp::MultiChannel.WUC(x = data[,dim], y = weight, k = c(min_level[dim]:kstar))
      result$size = as.vector(table(result$cluster))
      res <- structure(result,class = "Ckmeans.1d.dp")
      
      res.hist = Ckmeans.1d.dp::ahist(res, data = data[,dim], style = "midpoints", plot = FALSE)
      
      lines = unique(res.hist$breaks)
      grid_lines[[dim]] = lines[-c(1,length(lines))]
    }
  }

  else if(grid_method == "Sort+split"){
    # get grid lines
     grid = findgrid(cluster_info, length(unique(cluster_info$clusters)), nrow(data), 
                     ncol(data), min_level, max_level, "Density", cutoff, entropy, 
                     FALSE, nthread)
    # filter grid lines
    #grid_lines = lapply(grid_lines, function(i){ 
    #  mx = max(i) 
    #  return(i[-which(i == mx)])
    #  })
  }
  else if(grid_method == "DP"){
     # get grid lines
     grid = findgrid(cluster_info, length(unique(cluster_info$clusters)), 
                     nrow(data), ncol(data), min_level, max_level, "DP majority",
                     cutoff, entropy, FALSE, nthread)
     
  }
  else if(grid_method == "DP Compressed majority"){
     # get grid lines
     grid = findgrid(cluster_info, length(unique(cluster_info$clusters)), 
                     nrow(data), ncol(data), min_level, max_level, "DP compressed majority",
                     cutoff, entropy, FALSE, nthread)
     
  }
  else if(grid_method == "DP exact likelihood"){
     # get grid lines
     grid = findgrid(cluster_info, length(unique(cluster_info$clusters)), 
                     nrow(data), ncol(data), min_level, max_level, "DP exact likelihood", 
                     0, entropy, dim_reduction, nthread)
  }
  else if(grid_method == "DP approx likelihood 1-way"){
     # get grid lines
     #browser()
     grid = findgrid(cluster_info, length(unique(cluster_info$clusters)), 
                     nrow(data), ncol(data), min_level, max_level, "DP approx likelihood",
                     0, entropy, dim_reduction, nthread)
     #return(grid)
  }
  else if(grid_method == "DP approx likelihood 2-way"){
     # get grid lines
     #browser()
     grid = findgrid(cluster_info, length(unique(cluster_info$clusters)), 
                     nrow(data), ncol(data), min_level, max_level, "DP approx likelihood two way",
                     0, entropy, dim_reduction, nthread)
     #return(grid)
  }
  names(grid) <- c("lines", "table", "discr", "BIC", "purity", "lines_reduc", 
                   "table_reduc", "discr_reduc", "dimensional_bic", "BIC_reduc",
                   "purity_reduc", "line_removed", "upsilon", 
                   "distance_to_median", "distance_to_mean")
  
  names(grid$lines) <- colnames(data)
  if(FALSE){
  # discretize data
  discr_data = discretize_data(data, grid$lines)

  if(eval_method == "ARI"){
     # compute adjusted random index
     ndim_cluster_dist = discr_data[,1]
     if(ncol(discr_data) > 1){
        for(i in 2:ncol(discr_data)){
           ndim_cluster_dist = paste0(ndim_cluster_dist,",",discr_data[,i])
        }
     }
     #browser()
     block_ID = as.factor(ndim_cluster_dist)
     if(nlevels(block_ID) > 1){
        cluster_similarity = mclust::adjustedRandIndex(as.numeric(block_ID)-1, cluster_info$clusters)
     }
     else{
        cluster_similarity = 0
     }
  }
  else if(eval_method == "purity"){
     # compute adjusted random index
     ndim_cluster_dist = discr_data[,1]
     if(ncol(discr_data) > 1){
        for(i in 2:ncol(discr_data)){
           ndim_cluster_dist = paste0(ndim_cluster_dist,",",discr_data[,i])
        }
     }
     #browser()
     block_ID = as.factor(ndim_cluster_dist)
     #cluster_similarity = purity(cluster_info$clusters, as.numeric(block_ID)-1)
     #browser()
     cluster_similarity = ClusterPurity(as.numeric(block_ID)-1, cluster_info$clusters)
  }
  }
  
  # compute adjusted random index
  N = dim(data)[1]
  ndim = dim(data)[2]
  discr_data = grid$discr
  discr_data_after = grid$discr_reduc
  
  discr_data = matrix(unlist(discr_data), nrow=N, byrow = TRUE)
  if(dim_reduction){
     discr_data_after = matrix(unlist(discr_data_after), nrow=N, byrow = TRUE)
  }
  
  # browser()
  # compute adjusted random index for original data
  ndim_cluster_dist = discr_data[,1]
  if(ncol(discr_data) > 1){
     for(i in 2:ncol(discr_data)){
        ndim_cluster_dist = paste0(ndim_cluster_dist,",",discr_data[,i])
     }
  }
  block_ID = as.factor(ndim_cluster_dist)
  if(nlevels(block_ID) > 1){
     #cluster_similarity_o = mclust::adjustedRandIndex(as.numeric(block_ID)-1, cluster_info$clusters)
     cluster_similarity_o =  mclust::adjustedRandIndex(as.numeric(block_ID)-1, cluster_info$clusters)
  }
  else{
     cluster_similarity_o = 0
  }
  
  if(dim_reduction){
     # compute adjusted random index for removed data
     ndim_cluster_dist_r = discr_data_after[,1]
     if(ncol(discr_data_after) > 1){
        for(i in 2:ncol(discr_data_after)){
           ndim_cluster_dist_r = paste0(ndim_cluster_dist_r,",",discr_data_after[,i])
        }
     }
     block_ID = as.factor(ndim_cluster_dist_r)
     if(nlevels(block_ID) > 1){
        #cluster_similarity_r = mclust::adjustedRandIndex(as.numeric(block_ID)-1, cluster_info$clusters)
        cluster_similarity_r =  mclust::adjustedRandIndex(as.numeric(block_ID)-1, cluster_info$clusters)
     }
     else{
        cluster_similarity_r = 0
     }
  }
  else{
     cluster_similarity_r = 0
  }
  
  
  contig_table_o = matrix(unlist(grid$table), nrow=length(grid$table), byrow = TRUE)
  if(dim_reduction){
     contig_table_r = matrix(unlist(grid$table_reduc), nrow=length(grid$table_reduc), byrow = TRUE)
  }
  else{
     contig_table_r = 0
  }
  
  # calculate upsilon
  max_upsilon = (prod(apply(discr_data, 2, function(col) length(unique(col))))*(dim(data)[1]))/(2^(dim(data)[2]))
  u.res = hashed_upsilon(discr_data, TRUE)
  uval = u.res$statistic
  #urelt = uval / max_upsilon * (dim(data)[1])
  urelt = u.res$stat_relt
  pval = u.res$p.value
  #prelt = pchisq(urelt, df=1, lower.tail = FALSE, log.p = TRUE)
  prelt = u.res$p.value_relt
  
  # calculate p-val for upsilon
  #pval = 0
  #pval = pchisq(as.numeric(grid$upsilon), df = (dim(contig_table_o)[1]-1)*(dim(contig_table_o)[2]-1), 
  #              lower.tail = FALSE, log.p = TRUE)
  
  # print(cluster_similarity_o)
  # print(cluster_similarity_r)
  
  class = "GridOnClusters"
  # browser()
  
  # 1. Construct the base list with all the shared, standard outputs
  base_return <- list(
     D               = discr_data, 
     D_after         = discr_data_after, 
     grid            = grid$lines, 
     contig_table    = contig_table_o, 
     clabels         = cluster_info$clusters + 1, 
     dlabels         = as.numeric(block_ID) - 1, 
     dimensional_bic = grid$dimensional_bic, 
     csimilarity     = cluster_similarity_o, 
     BIC             = grid$BIC, 
     pval            = pval, 
     prelt           = prelt,
     upsilon         = uval, 
     upsilon_relt    = urelt, 
     cluster_method  = cluster_method, 
     upsilon_level   = u.res$upsilon_level, 
     add_noise       = cluster_info$add_noise,
     grid_method     = grid_method, 
     data            = data
  )
  
  # 2. Append the debug-specific elements if requested
  if (debug) {
     debug_elements <- list(
        grid_r              = grid$lines_reduc, 
        contig_table_r      = contig_table_r, 
        csimilarity_r       = cluster_similarity_r, 
        BIC_r               = grid$BIC_reduc,
        purity              = grid$purity, 
        purity_r            = grid$purity_reduc, 
        tsp                 = u.res$tsp, 
        tsm                 = u.res$tsm,
        distance_to_median  = grid$distance_to_median, 
        mean_dist_to_median = mean(unlist(grid$distance_to_median)),
        distance_to_mean    = grid$distance_to_mean, 
        mean_dist_to_mean   = mean(unlist(grid$distance_to_mean))
     )
     
     base_return <- c(base_return, debug_elements)
  }
  
  # 3. Apply the class structure once at the end
  res <- structure(base_return, class = class)
}

#' Cluster Multivariate Data
#' 
#' The function obtains clusters from data using the given
#'   number of clusters, which may be a range.
#'   
#' @param data input continuous multivariate data
#' @param k the number(s) of clusters 
#' @param method the method for clustering
#' @param noise adding jitter noise to the data or not
#' @param min_pts an integer to specify the minimum number of points inside a cluster.
#' @param nthread an integer to specify the number of threads
#'
#' @export
cluster = function(data, k, method, noise, min_pts, nthread){
  if(method == "Ball+BIC"){
    if(all(is.finite(k))){
      
       if(noise == TRUE){
          ### TODO: to be complate the part if add noise is needed
          # use noisy Mclust
          #mclust.res = Call_Noisy_Mclust(data, k)
       }else{
          VII.res = ballclust(data, G = k, threads = nthread,
                              min_pts = min_pts, verbose = FALSE)
          
          # add additional information to Mclust (non noisy data)
          if(!is.null(VII.res)){
             VII.res$isNoisy = FALSE  
          }
          else{
             ### TODO: to be complate the part if data too dense
             #mclust.res = Call_Noisy_Mclust(data, k)
          }
       }

      cluster_info = list(
        clusters = as.vector(as.numeric(as.factor(VII.res$classification))-1),
        data = as.matrix(data), method = "Ball+BIC", add_noise = VII.res$isNoisy)
        
    } else {
      ### TODO: to be complate the part if K is infinite
       
      range_k = c(2:8)
      
      # use noisy Mclust
      mclust.res = Call_Noisy_Mclust(data, range_k)
      
      # find the number of best clusters
      BIC_collect = mclust.res$BIC[,1]
      bestk = as.numeric(names(which.max(BIC_collect)))
      ogk = range_k
      while(bestk == max(range_k)){
        
        range_k = c((max(range_k)+1):min((max(range_k)*2), nrow(data)))
        ogk = c(ogk, range_k)
        
        mclust.res = Call_Noisy_Mclust(data, range_k)
        
        BIC_collect = c(BIC_collect, mclust.res$BIC[,1])  
        bestk = as.numeric(names(which.max(BIC_collect)))
        
      }
      
      # # let the user know that the user has reached the best k
      # message(paste0("Reached best k=",bestk,
      #                ". Re-clustering data using the best k."))
      
      # re-cluster using best k
      mclust.res = Call_Noisy_Mclust(data, bestk)
      
      cluster_info = list(
        clusters = as.vector(as.numeric(as.factor(mclust.res$classification))-1),
        data = as.matrix(data), method = method, add_noise = mclust.res$isNoisy)
      
    }
  } else if(method == "mclust_VII"){
     if(all(is.finite(k))){
        
        if(noise == TRUE){
           # use noisy Mclust
           mclust.res = Call_Noisy_Mclust(data, k)
        }else{
           mclust.res = mclust::Mclust(data, G = k, modelNames = "VII", 
                                       verbose = FALSE)
           
           # add additional information to Mclust (non noisy data)
           if(!is.null(mclust.res)){
              mclust.res$isNoisy = FALSE  
           }
           else{
              mclust.res = Call_Noisy_Mclust(data, k)
           }
        }
        
        cluster_info = list(
           clusters = as.vector(as.numeric(as.factor(mclust.res$classification))-1),
           data = as.matrix(data), method = "Ball+BIC", add_noise = mclust.res$isNoisy)
        
     } else {
        
        range_k = c(2:8)
        
        # use noisy Mclust
        mclust.res = Call_Noisy_Mclust(data, range_k)
        
        # find the number of best clusters
        BIC_collect = mclust.res$BIC[,1]
        bestk = as.numeric(names(which.max(BIC_collect)))
        ogk = range_k
        while(bestk == max(range_k)){
           
           range_k = c((max(range_k)+1):min((max(range_k)*2), nrow(data)))
           ogk = c(ogk, range_k)
           
           mclust.res = Call_Noisy_Mclust(data, range_k)
           
           BIC_collect = c(BIC_collect, mclust.res$BIC[,1])  
           bestk = as.numeric(names(which.max(BIC_collect)))
           
        }
        
        # # let the user know that the user has reached the best k
        # message(paste0("Reached best k=",bestk,
        #                ". Re-clustering data using the best k."))
        
        # re-cluster using best k
        mclust.res = Call_Noisy_Mclust(data, bestk)
        
        cluster_info = list(
           clusters = as.vector(as.numeric(as.factor(mclust.res$classification))-1),
           data = as.matrix(data), method = method, add_noise = mclust.res$isNoisy)
        
     }
  } else if(FALSE && method == "Ball+Tele+BIC"){
     # require(Telescope)
     # require(mclust)
     if(all(is.finite(k))){
        # tel = Telescope::telescope(data, TRUE)
        tele_tree = hc(data,modelName = "VII")
        tele_tree[,] = t(tele_tree) # t(my_tree)
        if(noise == TRUE){
           # use noisy Mclust
           mclust.res = Call_Noisy_Mclust(data, k) # , initialization = tele_tree)
        }else{
           mclust.res = mclust::Mclust(data, G = k, modelNames = "VII", 
                                       initialization = tele_tree, verbose = FALSE)
           
           # add additional information to Mclust (non noisy data)
           if(!is.null(mclust.res)){
              mclust.res$isNoisy = FALSE  
           }
        }
        
        cluster_info = list(
           clusters = as.vector(as.numeric(as.factor(mclust.res$classification))-1),
           data = as.matrix(data), method = "Ball+Tele+BIC")
        
     } else {
        
        range_k = c(2:8)
        
        # use noisy Mclust
        mclust.res = Call_Noisy_Mclust(data, range_k)
        
        # find the number of best clusters
        BIC_collect = mclust.res$BIC[,1]
        bestk = as.numeric(names(which.max(BIC_collect)))
        ogk = range_k
        while(bestk == max(range_k)){
           
           range_k = c((max(range_k)+1):min((max(range_k)*2), nrow(data)))
           ogk = c(ogk, range_k)
           
           mclust.res = Call_Noisy_Mclust(data, range_k)
           
           BIC_collect = c(BIC_collect, mclust.res$BIC[,1])  
           bestk = as.numeric(names(which.max(BIC_collect)))
           
        }
        
        # # let the user know that the user has reached the best k
        # message(paste0("Reached best k=",bestk,
        #                ". Re-clustering data using the best k."))
        
        # re-cluster using best k
        mclust.res = Call_Noisy_Mclust(data, bestk)
        
        cluster_info = list(
           clusters = as.vector(as.numeric(as.factor(mclust.res$classification))-1),
           data = as.matrix(data), method = method)
        
     }
     
  } 
   else if(method == "kmeans+silhouette"){
    # is k a single number or a range
    if(length(k) == 1 && is.finite(k)){
      
      # randomly generate centers
      centers = dqsample(which(!duplicated(data)), k)
      
      # get cluster information for 'k'
      kmeans.res = kmeans(data, centers = as.matrix(data[centers,]), 
                            iter.max = 100)
      
      # only keep cluster centers and labels
      cluster_info = list(#centers = as.matrix(kmeans.res$centers),
        clusters = as.vector(kmeans.res$cluster)-1,
        data = as.matrix(data), method = method)
      
    } else if(length(k) == 1 && !is.finite(k)) { # else if k is set to Inf
      
      data_dist = dist(data) # distance matrix for data
      
      # compute cluster info and silhouette score for the cluster range k, 
      # adaptively increasing k until optimal or each sample is placed in their 
      # own cluster
      
      # find unique number of samples
      unq = sum(!duplicated(data))
      if(unq > 8){
        range_k = c(2:8)  
      } else {
        range_k = c(2:unq)
      }
      
      # compute cluster info for original k range
      cluster_info = lapply(range_k, function(i){
        centers = dqsample(which(!duplicated(data)), i)
        data_clust = kmeans(data, centers = as.matrix(data[centers,]),
                            iter.max = 100)
        return(list(data_clust$cluster, data_clust$centers, 
                    mean(silhouette(data_clust$cluster, dist=data_dist)[,3])))
      })
      
      # find the max silhouette
      silhouette_scr = unlist(lapply(cluster_info, function(x){
        return(x[[3]])
      }), use.names = FALSE)
      max_silhouette = max(which.max(silhouette_scr)) # take the bigger k out of those with equal silhouette scores
      
      # while best k is equal to max k, keep computing scores
      bestk = range_k[max_silhouette]
      ogk = range_k
      while(bestk == max(range_k) && unq > 8){
        
        range_k = c((max(range_k)+1):min((max(range_k)*2), nrow(data)))
        ogk = c(ogk, range_k)
        
        # compute cluster info for original k range
        n_cluster_info = lapply(range_k, function(i){
          centers = dqsample(which(!duplicated(data)), i)
          data_clust = kmeans(data, centers = as.matrix(data[centers,]),
                              iter.max = 100)
          return(list(data_clust$cluster, data_clust$centers, 
                      mean(silhouette(data_clust$cluster, dist=data_dist)[,3])))
        })
        
        # append new cluster info to original
        cluster_info = c(cluster_info, n_cluster_info)
        
        # find the max silhouette
        silhouette_scr = unlist(lapply(cluster_info, function(x){
          return(x[[3]])
        }), use.names = FALSE)
        max_silhouette = max(which.max(silhouette_scr)) # take the bigger k out of those with equal silhouette scores
        
        # find best k
        bestk = ogk[max_silhouette]
      }
      
      # only keep cluster centers and labels for the 'k' with max silhouette
      cluster_info = list(#centers = as.matrix(cluster_info[[max_silhouette]][[2]]),
        clusters = as.vector(cluster_info[[max_silhouette]][[1]]-1),
        data = as.matrix(data), cluster_method = method)
      
    } else { # else k is a vector
      
      data_dist = dist(data) # distance matrix for data
      
      # compute cluster info and silhouette score for the cluster range k
      cluster_info = lapply(k, function(i){
        centers = dqsample(which(!duplicated(data)), i)
        data_clust = kmeans(data, centers = as.matrix(data[centers,]),
                            iter.max = 100)
        return(list(data_clust$cluster, data_clust$centers, 
                    mean(silhouette(data_clust$cluster, dist=data_dist)[,3])))
      })
      
      # find the max silhouette
      silhouette_scr = unlist(lapply(cluster_info, function(x){
        return(x[[3]])
      }), use.names = FALSE)
      max_silhouette = max(which.max(silhouette_scr)) # take the bigger k out of those with equal silhouette scores
      
      # only keep cluster centers and labels for the 'k' with max silhouette
      cluster_info = list(#centers = as.matrix(cluster_info[[max_silhouette]][[2]]),
        clusters = as.vector(cluster_info[[max_silhouette]][[1]]-1),
        data = as.matrix(data), cluster_method = method)
      
    }
  }
   else if(method == "PAM"){
      if(length(k) == 1 && is.finite(k)){
         # pre-cluster the data using partition around medoids (PAM)
         cluster_label = pam(x=data, diss = FALSE, metric = "euclidean", k = k)$clustering
      }
      else{
         # error
      }
      
      cluster_info = list(
         clusters = as.vector(cluster_label), data = as.matrix(data), method = method)
   }
   else if(FALSE && method == "telescope"){
      # require(Telescope)
      res <- NULL # Telescope::telescope(data,0)
      cluster_info = list(
         clusters = res$c_label, data = as.matrix(data), method = method)
   }
  return(cluster_info)
}

# for internal use
# takes n-dimensional 'data' and clusters it into 'k' groups using Mclust
# adding noise whenever necessary
Call_Noisy_Mclust = function(data, k){
  
  mclust.res = mclust::Mclust(data, G = k, modelNames = "VII", 
                              verbose = FALSE)
  
  # add additional information to Mclust (non noisy data)
  if(!is.null(mclust.res)){
    mclust.res$isNoisy = FALSE  
  }
  
  # Mclust can return NULL result for sparse discrete data
  amt = 0.5
  noisy_data = data # don't change the original input
  while(is.null(mclust.res)){
    
    # add a little bit of noise to enable clustering
    for(j in 1:ncol(noisy_data)){
      noisy_data[,j] = jitter(as.numeric(noisy_data[,j]), amount = amt)
    }
    
    mclust.res = mclust::Mclust(noisy_data, G = k, modelNames = "VII", 
                                verbose = FALSE)
    
    # add additional information to Mclust (noisy data)
    if(!is.null(mclust.res)){
      mclust.res$isNoisy = TRUE
    }
    
    # double the amount of noise for the next iteration (if any)
    amt = amt*2
    
  }
  
  if(mclust.res$isNoisy){
    warning(paste0("Mclust could not cluster 'data', jitter noise 
            (factor=1, amount=",(amt/2),") was added."))
  }
  
  return(mclust.res)
  
}

# for internal use
# takes n-dimesional 'data' and gridlines to quantify each dimension
discretize_data = function(data, gridlines){
  
  # use gridlines for each dimension
  for(i in 1:ncol(data)){
    
    if(length(unique(gridlines))==0){
      discr = rep(1, nrow(data))
    }else{
      discr = rep(length(gridlines[[i]])+1, nrow(data))
      gridlines[[i]] = sort(gridlines[[i]])
      for(j in 1:length(gridlines[[i]])){ # determine discretization levels
        if(j == 1) {
          discr[data[,i] < gridlines[[i]][j]] = 1
        } else {
          discr[data[,i] < gridlines[[i]][j] & data[,i] >= gridlines[[i]][j-1]] = j
        }
      }
    }
    data[,i] = discr
  }
  
  return(data)
}

# test-GridOnClusters.R
#
# tests discretize.jointly function under the perameter (cluster_method == "Ball+BIC" 
# and grid_method == "DP Compressed likelihood")
# Created by: Jiandong Wang, Sajal Kumar and Dr. Mingzhou (Joe) Song
# Date Created: 16th August, 2022

library(testthat)
library(FunChisq)
library(GridOnClusters)
library(cluster)
library(dqrng)
library(mclust)

context("Testing Ball+BIC & DP approx likelihood 1-way")

test_that("Testing discretize.jointly (\"Ball+BIC\")(\"DP approx likelihood 1-way\")", {
   
   # test 1
   # y = f(x)
   # z = f(x)
   # k = constant
   
   cluster_method <- "Ball+BIC"
   grid_method <- "DP approx likelihood 1-way"
   
   dqset.seed(123)
   set.seed(123)
   
   x = dqrnorm(100, mean=5, sd=1)
   y = sin(x)
   z = cos(x)
   data = cbind(x, y, z)
   discr = discretize.jointly(
      data, k=3, cluster_method=cluster_method,
      grid_method = grid_method, min_level = 1)
   
   # test marginal levels
   expect_equivalent(length(unique(discr$D[,1])), 3)
   expect_equivalent(length(unique(discr$D[,2])), 2)
   expect_equivalent(length(unique(discr$D[,3])), 3)
   
   # test marginal distribution
   expect_equivalent(table(discr$D[,1]), as.table(c(24, 39, 37)))
   expect_equivalent(table(discr$D[,2]), as.table(c(46, 54)))
   expect_equivalent(table(discr$D[,3]), as.table(c(23, 43, 34)))
   
   # test 2d joint distributions
   dim12 = table(discr$D[,1], discr$D[,2])
   expect_equivalent(dim12, as.table(matrix(c(7, 17,
                                              39, 0,
                                              0, 37),
                                            nrow=3, ncol=2, byrow = T)))
   
   dim13 = table(discr$D[,1], discr$D[,3])
   expect_equivalent(dim13, as.table(matrix(c(23, 1, 0,
                                              0, 39, 0,
                                              0, 3, 34),
                                            nrow=3, ncol=3, byrow = T)))
   
   dim23 = table(discr$D[,2], discr$D[,3])
   expect_equivalent(dim23, as.table(matrix(c(7, 39, 0,
                                              16, 4, 34),
                                            nrow=2, ncol=3, byrow = T)))
   
   # test ARI score
   #expect_equivalent(round(discr$csimilarity, digits = 3), 0.949)
   expect_equivalent(round(discr$csimilarity, digits = 3), 0.89)
   
   # test scale
   discr.scale = discretize.jointly(
      data, k=3, cluster_method=cluster_method,
      grid_method = grid_method, min_level = 1, scale = TRUE)
   
   # test marginal levels
   expect_equivalent(length(unique(discr.scale$D[,1])), 3)
   expect_equivalent(length(unique(discr.scale$D[,2])), 2)
   expect_equivalent(length(unique(discr.scale$D[,3])), 3)
   
   # test marginal distribution
   expect_equivalent(table(discr.scale$D[,1]), as.table(c(24, 39, 37)))
   expect_equivalent(table(discr.scale$D[,2]), as.table(c(46, 54)))
   expect_equivalent(table(discr.scale$D[,3]), as.table(c(23, 43, 34)))
   
   # test 2d joint distributions
   dim12 = table(discr.scale$D[,1], discr.scale$D[,2])
   expect_equivalent(dim12, as.table(matrix(c(7, 17,
                                              39, 0,
                                              0, 37),
                                            nrow=3, ncol=2, byrow = T)))
   
   dim13 = table(discr.scale$D[,1], discr.scale$D[,3])
   expect_equivalent(dim13, as.table(matrix(c(23, 1, 0,
                                              0, 39, 0,
                                              0, 3, 34),
                                            nrow=3, ncol=3, byrow = T)))
   
   dim23 = table(discr.scale$D[,2], discr.scale$D[,3])
   expect_equivalent(dim23, as.table(matrix(c(7, 39, 0,
                                              16, 4, 34),
                                            nrow=2, ncol=3, byrow = T)))
   
   # test ARI score
   #expect_equivalent(round(discr.scale$csimilarity, digits = 3), 0.949)
   expect_equivalent(round(discr$csimilarity, digits = 3), 0.89)
   
   # test Scaled Upsilon score
   expect_equivalent(round(discr$upsilon_relt, digits = 3), 1.881)
   # test 2
   # y = f(x)
   # z = f(x)
   # k = variable (determined by silhouette)
   dqset.seed(321)
   set.seed(321)
   
   x = dqrnorm(n = 100, mean=10, sd=2)
   y = log(x)
   z = tan(x)
   data = cbind(x, y, z)
   discr = discretize.jointly(data, k=c(3:10), cluster_method=cluster_method,
                              grid_method = grid_method, min_pts = 3, min_level = 1)
   
   # test marginal levels
   expect_equivalent(length(unique(discr$D[,1])), 3)
   expect_equivalent(length(unique(discr$D[,2])), 3)
   expect_equivalent(length(unique(discr$D[,3])), 4)
   
   # test marginal distribution
   expect_equivalent(table(discr$D[,1]), as.table(c(22, 34, 44)))
   expect_equivalent(table(discr$D[,2]), as.table(c(22, 34, 44)))
   expect_equivalent(table(discr$D[,3]), as.table(c(12, 31, 30, 27)))
   
   # test 2d joint distributions
   dim12 = table(discr$D[,1], discr$D[,2])
   expect_equivalent(dim12, as.table(matrix(c(22,  0,  0, 
                                               0, 34,  0, 
                                               0,  0, 44),
                                            nrow=3, ncol=3, byrow = T)))
   dim13 = table(discr$D[,1], discr$D[,3])
   expect_equivalent(dim13, as.table(matrix(c(4,  11,  3, 4,
                                              0,  0, 22, 12,
                                              8, 20,  5, 11),
                                            nrow=3, ncol=4, byrow = T)))
   dim23 = table(discr$D[,2], discr$D[,3])
   expect_equivalent(dim23, as.table(matrix(c(4,  11,  3, 4,
                                              0,  0, 22, 12,
                                              8, 20,  5, 11),
                                            nrow=3, ncol=4, byrow = T)))
   # test ARI score
   #expect_equivalent(round(discr$csimilarity, digits = 3), 0.857)
   expect_equivalent(round(discr$csimilarity, digits = 3), 0.784)
   
   # test Scaled Upsilon score
   expect_equivalent(round(discr$upsilon_relt, digits = 3), 0.812)
   # test 3
   # y != f(x)
   # z = f(x, y)
   # k = variable (determined by silhouette)
   dqset.seed(1234)
   set.seed(1234)
   
   x = dqrexp(n=50, rate = 0.6)
   y = dqrnorm(50, mean=2, sd=0.5)
   z = sin(x) + cos(y)
   data = cbind(x, y, z)
   discr = discretize.jointly(
      data, k=c(3:10), min_level = 2, cluster_method=cluster_method,
      grid_method = grid_method)
   
   # test marginal levels
   expect_equivalent(length(unique(discr$D[,1])), 3)
   expect_equivalent(length(unique(discr$D[,2])), 2)
   expect_equivalent(length(unique(discr$D[,3])), 3)
   
   # test marginal distribution
   expect_equivalent(table(discr$D[,1]), as.table(c(27, 14, 9)))
   expect_equivalent(table(discr$D[,2]), as.table(c(25, 25)))
   expect_equivalent(table(discr$D[,3]), as.table(c(8, 20, 22)))
   
   # test 2d joint distributions
   dim12 = table(discr$D[,1], discr$D[,2])
   expect_equivalent(
     dim12, as.table(matrix(c(15, 12,
                              7, 7,
                              3, 6),
                            nrow=3, ncol=2, byrow = T)))
   
   dim13 = table(discr$D[,1], discr$D[,3])
   expect_equivalent(
     dim13, as.table(matrix(c(0, 13, 14,
                              0,  7,  7,
                              8,  0, 1),
                            nrow=3, ncol=3, byrow = T)))
   
   dim23 = table(discr$D[,2], discr$D[,3])
   expect_equivalent(
     dim23, as.table(matrix(c(2,  2, 21,
                              6, 18,  1),
                            nrow=2, ncol=3, byrow = T)))
   
   # test ARI score
   expect_equivalent(round(discr$csimilarity, digits = 3), 0.742)
   
   # test Scaled Upsilon score
   expect_equivalent(round(discr$upsilon_relt, digits = 3), 0.775)
   
   
   #### use mclust to cluster
   cluster_method <- "mclust_VII"
   grid_method <- "DP approx likelihood 1-way"
   
   dqset.seed(123)
   set.seed(123)
   
   x = dqrnorm(100, mean=5, sd=1)
   y = sin(x)
   z = cos(x)
   data = cbind(x, y, z)
   discr = discretize.jointly(
      data, k=3, cluster_method=cluster_method,
      grid_method = grid_method, min_level = 1)
   
   # test marginal levels
   expect_equivalent(length(unique(discr$D[,1])), 4)
   expect_equivalent(length(unique(discr$D[,2])), 2)
   expect_equivalent(length(unique(discr$D[,3])), 3)
   
   # test marginal distribution
   expect_equivalent(table(discr$D[,1]), as.table(c(24, 39, 33, 4)))
   expect_equivalent(table(discr$D[,2]), as.table(c(46, 54)))
   expect_equivalent(table(discr$D[,3]), as.table(c(23, 43, 34)))
   
   # test 2d joint distributions
   dim12 = table(discr$D[,1], discr$D[,2])
   expect_equivalent(dim12, as.table(matrix(c(7, 17,
                                              39, 0,
                                              0, 33,
                                              0, 4),
                                            nrow=4, ncol=2, byrow = T)))
   
   dim13 = table(discr$D[,1], discr$D[,3])
   expect_equivalent(dim13, as.table(matrix(c(23, 1, 0,
                                              0, 39, 0,
                                              0, 0, 33,
                                              0, 3, 1),
                                            nrow=4, ncol=3, byrow = T)))
   
   dim23 = table(discr$D[,2], discr$D[,3])
   expect_equivalent(dim23, as.table(matrix(c(7, 39, 0,
                                              16, 4, 34),
                                            nrow=2, ncol=3, byrow = T)))
   
   # test ARI score
   #expect_equivalent(round(discr$csimilarity, digits = 3), 0.949)
   expect_equivalent(round(discr$csimilarity, digits = 3), 0.89)
   
   # test scale
   discr.scale = discretize.jointly(
      data, k=3, cluster_method=cluster_method,
      grid_method = grid_method, min_level = 1, scale = TRUE)
   
   # test marginal levels
   expect_equivalent(length(unique(discr.scale$D[,1])), 4)
   expect_equivalent(length(unique(discr.scale$D[,2])), 3)
   expect_equivalent(length(unique(discr.scale$D[,3])), 3)
   
   # test marginal distribution
   expect_equivalent(table(discr.scale$D[,1]), as.table(c(24, 39, 32, 5)))
   expect_equivalent(table(discr.scale$D[,2]), as.table(c(46, 47, 7)))
   expect_equivalent(table(discr.scale$D[,3]), as.table(c(23, 43, 34)))
   
   # test 2d joint distributions
   dim12 = table(discr.scale$D[,1], discr.scale$D[,2])
   expect_equivalent(dim12, as.table(matrix(c(7, 15, 2,
                                              39, 0, 0,
                                              0, 32, 0,
                                              0, 0, 5),
                                            nrow=4, ncol=3, byrow = T)))
   
   dim13 = table(discr.scale$D[,1], discr.scale$D[,3])
   expect_equivalent(dim13, as.table(matrix(c(23, 1, 0,
                                              0, 39, 0,
                                              0, 0, 32,
                                              0, 3, 2),
                                            nrow=4, ncol=3, byrow = T)))
   
   dim23 = table(discr.scale$D[,2], discr.scale$D[,3])
   expect_equivalent(dim23, as.table(matrix(c(7, 39, 0,
                                              15, 0, 32,
                                              1, 4, 2),
                                            nrow=3, ncol=3, byrow = T)))
   
   # test ARI score
   #expect_equivalent(round(discr.scale$csimilarity, digits = 3), 0.949)
   expect_equivalent(round(discr$csimilarity, digits = 3), 0.89)
   
   # test Scaled Upsilon score
   expect_equivalent(round(discr$upsilon_relt, digits = 3), 1.876)
   # test 2
   # y = f(x)
   # z = f(x)
   # k = variable (determined by silhouette)
   dqset.seed(321)
   set.seed(321)
   
   x = dqrnorm(n = 100, mean=10, sd=2)
   y = log(x)
   z = tan(x)
   data = cbind(x, y, z)
   discr = discretize.jointly(data, k=c(3:10), cluster_method=cluster_method,
                              grid_method = grid_method, min_level = 1)
   
   # test marginal levels
   expect_equivalent(length(unique(discr$D[,1])), 6)
   expect_equivalent(length(unique(discr$D[,2])), 6)
   expect_equivalent(length(unique(discr$D[,3])), 3)
   
   # test marginal distribution
   expect_equivalent(table(discr$D[,1]), as.table(c(19, 19, 18, 12, 21, 11)))
   expect_equivalent(table(discr$D[,2]), as.table(c(19, 19, 18, 12, 21, 11)))
   expect_equivalent(table(discr$D[,3]), as.table(c(35, 28, 37)))
   
   # test 2d joint distributions
   dim12 = table(discr$D[,1], discr$D[,2])
   expect_equivalent(dim12, as.table(matrix(c(19,  0,  0,  0,  0,  0,
                                              0, 19,  0,  0,  0,  0,
                                              0,  0, 18,  0,  0,  0,
                                              0,  0,  0, 12,  0,  0,
                                              0,  0,  0,  0, 21,  0,
                                              0,  0,  0,  0,  0, 11),
                                            nrow=6, ncol=6, byrow = T)))
   dim13 = table(discr$D[,1], discr$D[,3])
   expect_equivalent(dim13, as.table(matrix(c(11,  2,  6,
                                              0, 19,  0,
                                              0,  0, 18,
                                              7,  0,  5,
                                              16,  5,  0,
                                              1,  2,  8),
                                            nrow=6, ncol=3, byrow = T)))
   dim23 = table(discr$D[,2], discr$D[,3])
   expect_equivalent(dim23, as.table(matrix(c(11,  2,  6,
                                              0, 19,  0,
                                              0,  0, 18,
                                              7,  0,  5,
                                              16,  5,  0,
                                              1,  2,  8),
                                            nrow=6, ncol=3, byrow = T)))
   
   # test ARI score
   #expect_equivalent(round(discr$csimilarity, digits = 3), 0.857)
   expect_equivalent(round(discr$csimilarity, digits = 3), 0.58)
   
   # test Scaled Upsilon score
   expect_equivalent(round(discr$upsilon_relt, digits = 3), 0.929)
   # test 3
   # y != f(x)
   # z = f(x, y)
   # k = variable (determined by silhouette)
   dqset.seed(1234)
   set.seed(1234)
   
   x = dqrexp(n=50, rate = 0.6)
   y = dqrnorm(50, mean=2, sd=0.5)
   z = sin(x) + cos(y)
   data = cbind(x, y, z)
   discr = discretize.jointly(
      data, k=c(3:10), min_level = 2, cluster_method=cluster_method,
      grid_method = grid_method)
   
   # test marginal levels
   expect_equivalent(length(unique(discr$D[,1])), 2)
   expect_equivalent(length(unique(discr$D[,2])), 2)
   expect_equivalent(length(unique(discr$D[,3])), 3)
   
   # test marginal distribution
   expect_equivalent(table(discr$D[,1]), as.table(c(27, 23)))
   expect_equivalent(table(discr$D[,2]), as.table(c(32, 18)))
   expect_equivalent(table(discr$D[,3]), as.table(c(15, 23, 12)))
   
   # test 2d joint distributions
   dim12 = table(discr$D[,1], discr$D[,2])
   expect_equivalent(
      dim12, as.table(matrix(c(19, 8,
                               13, 10),
                             nrow=2, ncol=2, byrow = T)))
   
   dim13 = table(discr$D[,1], discr$D[,3])
   expect_equivalent(
      dim13, as.table(matrix(c(5, 13, 9,
                               10, 10, 3),
                             nrow=2, ncol=3, byrow = T)))
   
   dim23 = table(discr$D[,2], discr$D[,3])
   expect_equivalent(
      dim23, as.table(matrix(c(2, 18, 12,
                               13, 5, 0),
                             nrow=2, ncol=3, byrow = T)))
   
   # test ARI score
   expect_equivalent(round(discr$csimilarity, digits = 3), 0.546)
   
   # test Scaled Upsilon score
   expect_equivalent(round(discr$upsilon_relt, digits = 3), 0.358)
   
   
   # test 4
   # y = f(x)
   # z = f(x)
   # k = fixed
   # using an alternate clustering strategy
   dqset.seed(2468)
   set.seed(2468)
   
   x = dqrnorm(n = 1000, mean = 10, sd = 2)
   y = sin(x)
   z = cos(y)
   data = cbind(x, y, z)
   
   #### use PAM to cluster
   alt.cluster = pam(x = data, k = 5, diss = FALSE, metric = "euclidean", cluster.only = TRUE)
   
   discr = discretize.jointly(data = data, cluster_label = alt.cluster,
                              grid_method = grid_method, min_level = 1)
   
   # test marginal levels
   expect_equivalent(length(unique(discr$D[,1])), 5)
   expect_equivalent(length(unique(discr$D[,2])), 5)
   expect_equivalent(length(unique(discr$D[,3])), 4)
   
   # test marginal distribution
   expect_equivalent(table(discr$D[,1]), as.table(c(93, 195, 203, 320, 189)))
   expect_equivalent(table(discr$D[,2]), as.table(c(328, 25, 354, 149, 144)))
   expect_equivalent(table(discr$D[,3]), as.table(c(336, 222, 120, 322)))
   
   # test 2d joint distributions
   dim12 = table(discr$D[,1], discr$D[,2])
   expect_equivalent(dim12, as.table(matrix(c(12, 3, 41, 37, 0,
                                              0,  0,  0, 74, 121,
                                              0, 17,186, 0,  0, 
                                              315, 5, 0, 0,  0,
                                              1, 0, 127, 38, 23),
                                            nrow=5, ncol=5, byrow = T)))
   
   dim13 = table(discr$D[,1], discr$D[,3])
   expect_equivalent(dim13, as.table(matrix(c(6,   28,  20,  39,
                                              120, 46,  29,   0,
                                              0,    0,  34, 169,
                                              187,128,   5,   0,
                                              23,  20,  32, 114),
                                            nrow=5, ncol=4, byrow = T)))
   
   dim23 = table(discr$D[,2], discr$D[,3])
   expect_equivalent(dim23, as.table(matrix(c(194, 134,   0,   0,
                                              0,     0,  25,   0,
                                              0,     0,  34, 320,
                                              0,    86,  61,   2,
                                              142,   2,   0,   0),
                                            nrow=5, ncol=4, byrow = T)))
   
   # test ARI score
   expect_equivalent(round(discr$csimilarity, digits = 3), 0.61)
   
   # test Scaled Upsilon score
   expect_equivalent(round(discr$upsilon_relt, digits = 3), 0.753)
})

# test-GridOnClusters.R
#
# tests discretize.jointly function
# Created by: Jiandong Wang, Sajal Kumar and Dr. Mingzhou (Joe) Song
# Date Created: 9th March, 2020

library(testthat)
library(FunChisq)
library(GridOnClusters)
library(cluster)
library(dqrng)

context("Testing kmeans+silhouette & Sort+split")

test_that("Testing discretize.jointly (\"kmeans+silhouette\")(\"Sort+split\")", {

  # test 1
  # y = f(x)
  # z = f(x)
  # k = constant

  cluster_method <- "kmeans+silhouette"
  grid_method <- "Sort+split"
  
  dqset.seed(123)
  x = dqrnorm(100, mean=5, sd=1)
  y = sin(x)
  z = cos(x)
  data = cbind(x, y, z)
  discr = discretize.jointly(
     data, k=3, cluster_method=cluster_method,
     grid_method = grid_method, min_level = 1)

  # test marginal levels
  expect_equivalent(length(unique(discr$D[,1])), 3)
  expect_equivalent(length(unique(discr$D[,2])), 2)
  expect_equivalent(length(unique(discr$D[,3])), 3)

  # test marginal distribution
  expect_equivalent(table(discr$D[,1]), as.table(c(23, 45, 32)))
  expect_equivalent(table(discr$D[,2]), as.table(c(40, 60)))
  expect_equivalent(table(discr$D[,3]), as.table(c(26, 42, 32)))

  # test 2d joint distributions
  dim12 = table(discr$D[,1], discr$D[,2])
  expect_equivalent(dim12, as.table(matrix(c(5, 18,
                                             35,10,
                                             0, 32),
                                           nrow=3, ncol=2, byrow = T)))

  dim13 = table(discr$D[,1], discr$D[,3])
  expect_equivalent(dim13, as.table(matrix(c(22,  1,  0,
                                             4,  38,  3,
                                             0,   3, 29),
                                           nrow=3, ncol=3, byrow = T)))

  dim23 = table(discr$D[,2], discr$D[,3])
  expect_equivalent(dim23, as.table(matrix(c(9, 31, 0,
                                             17, 11, 32),
                                           nrow=2, ncol=3, byrow = T)))

  # test ARI score
  #expect_equivalent(round(discr$csimilarity, digits = 3), 1)
  expect_equivalent(round(discr$csimilarity, digits = 3), 0.664)
  
  # test 2
  # y = f(x)
  # z = f(x)
  # k = variable (determined by silhouette)
  dqset.seed(321)

  x = dqrnorm(n = 100, mean=10, sd=2)
  y = log(x)
  z = tan(x)
  data = cbind(x, y, z)
  discr = discretize.jointly(data, k=c(3:10), cluster_method=cluster_method,
                             grid_method = grid_method, min_level = 1)

  # test marginal levels
  expect_equivalent(length(unique(discr$D[,1])), 9)
  expect_equivalent(length(unique(discr$D[,2])), 9)
  expect_equivalent(length(unique(discr$D[,3])), 6)

  # test marginal distribution
  expect_equivalent(table(discr$D[,1]), as.table(c(9, 10, 39,  3,  1,  1,  5, 22, 10)))
  expect_equivalent(table(discr$D[,2]), as.table(c(9, 10, 39,  3,  1,  1,  5, 22, 10)))
  expect_equivalent(table(discr$D[,3]), as.table(c(1, 1, 7, 30, 59, 2)))

  # test 2d joint distributions
  dim12 = table(discr$D[,1], discr$D[,2])
  expect_equivalent(dim12, as.table(matrix(c(9, 0, 0, 0, 0, 0, 0, 0, 0,
                                             0,10, 0, 0, 0, 0, 0, 0, 0,
                                             0, 0, 39, 0, 0, 0, 0, 0, 0,
                                             0, 0, 0, 3, 0, 0, 0, 0, 0,
                                             0, 0, 0, 0, 1, 0, 0, 0, 0,
                                             0, 0, 0, 0, 0, 1, 0, 0, 0,
                                             0, 0, 0, 0, 0, 0, 5, 0, 0,
                                             0, 0, 0, 0, 0, 0, 0,22, 0,
                                             0, 0, 0, 0, 0, 0, 0, 0,10),
                                           nrow=9, ncol=9, byrow = T)))
  dim13 = table(discr$D[,1], discr$D[,3])
  expect_equivalent(dim13, as.table(matrix(c(0,  0,  0,  1,  8,  0,
                                             0,  0,  2,  8,  0,  0,
                                             0,  0,  0,  2, 37,  0,
                                             0,  0,  0,  0,  1,  2,
                                             1,  0,  0,  0,  0,  0,
                                             0,  1,  0,  0,  0,  0,
                                             0,  0,  5,  0,  0,  0,
                                             0,  0,  0, 18,  4,  0,
                                             0,  0,  0,  1,  9,  0),
                                           nrow=9, ncol=6, byrow = T)))
  dim23 = table(discr$D[,2], discr$D[,3])
  expect_equivalent(dim23, as.table(matrix(c(0,  0,  0,  1,  8,  0,
                                             0,  0,  2,  8,  0,  0,
                                             0,  0,  0,  2, 37,  0,
                                             0,  0,  0,  0,  1,  2,
                                             1,  0,  0,  0,  0,  0,
                                             0,  1,  0,  0,  0,  0,
                                             0,  0,  5,  0,  0,  0,
                                             0,  0,  0, 18,  4,  0,
                                             0,  0,  0,  1,  9,  0),
                                           nrow=9, ncol=6, byrow = T)))

  # test ARI score
  #expect_equivalent(round(discr$csimilarity, digits = 3), 0.968)
  expect_equivalent(round(discr$csimilarity, digits = 3), 0.861)
  

  # test 3
  # y != f(x)
  # z = f(x, y)
  # k = variable (determined by silhouette)
  dqset.seed(1234)

  x = dqrexp(n=50, rate = 0.6)
  y = dqrnorm(50, mean=2, sd=0.5)
  z = sin(x) + cos(y)
  data = cbind(x, y, z)
  discr = discretize.jointly(
     data, k=c(3:10), min_level = 2, cluster_method=cluster_method,
     grid_method = grid_method)

  # test marginal levels
  expect_equivalent(length(unique(discr$D[,1])), 3)
  expect_equivalent(length(unique(discr$D[,2])), 2)
  expect_equivalent(length(unique(discr$D[,3])), 2)

  # test marginal distribution
  expect_equivalent(table(discr$D[,1]), as.table(c(29, 17, 4)))
  expect_equivalent(table(discr$D[,2]), as.table(c(34, 16)))
  expect_equivalent(table(discr$D[,3]), as.table(c(9, 41)))

  # test 2d joint distributions
  dim12 = table(discr$D[,1], discr$D[,2])
  expect_equivalent(dim12, as.table(matrix(c(21, 8,
                                             12, 5,
                                             1, 3),
                                           nrow=3, ncol=2, byrow = T)))

  dim13 = table(discr$D[,1], discr$D[,3])
  expect_equivalent(dim13, as.table(matrix(c(1, 28,
                                             4, 13,
                                             4, 0),
                                           nrow=3, ncol=2, byrow = T)))

  dim23 = table(discr$D[,2], discr$D[,3])
  expect_equivalent(dim23, as.table(matrix(c(4, 30,
                                             5, 11),
                                           nrow=2, ncol=2, byrow = T)))

  # test ARI score
  #expect_equivalent(round(discr$csimilarity, digits = 3), 0.821)
  expect_equivalent(round(discr$csimilarity, digits = 3), 0.534)
  
  # test 4
  # y = f(x)
  # z = f(x)
  # k = fixed
  # using an alternate clustering strategy
  dqset.seed(2468)

  x = dqrnorm(n = 1000, mean = 10, sd = 2)
  y = sin(x)
  z = cos(y)
  data = cbind(x, y, z)
  # use PAM to cluster
  alt.cluster = pam(x = data, k = 5, diss = FALSE, metric = "euclidean", cluster.only = TRUE)

  discr = discretize.jointly(data = data, cluster_label = alt.cluster,
                             grid_method = grid_method, min_level = 1)

  # test marginal levels
  expect_equivalent(length(unique(discr$D[,1])), 5)
  expect_equivalent(length(unique(discr$D[,2])), 3)
  expect_equivalent(length(unique(discr$D[,3])), 2)

  # test marginal distribution
  expect_equivalent(table(discr$D[,1]), as.table(c(93, 196, 203, 319, 189)))
  expect_equivalent(table(discr$D[,2]), as.table(c(334, 451, 215)))
  expect_equivalent(table(discr$D[,3]), as.table(c(439, 561)))

  # test 2d joint distributions
  dim12 = table(discr$D[,1], discr$D[,2])
  expect_equivalent(dim12, as.table(matrix(c(13, 61, 19,
                                             0, 40, 156,
                                             5, 198, 0,
                                             315, 4, 0,
                                             1, 148, 40),
                                           nrow=5, ncol=3, byrow = T)))

  dim13 = table(discr$D[,1], discr$D[,3])
  expect_equivalent(dim13, as.table(matrix(c(20, 73,
                                             138, 58,
                                             0, 203,
                                             250, 69,
                                             31, 158),
                                           nrow=5, ncol=2, byrow = T)))

  dim23 = table(discr$D[,2], discr$D[,3])
  expect_equivalent(dim23, as.table(matrix(c(261, 73,
                                             0, 451,
                                             178, 37),
                                           nrow=3, ncol=2, byrow = T)))

  # test ARI score
  #expect_equivalent(round(discr$csimilarity, digits = 3), 0.881)
  expect_equivalent(round(discr$csimilarity, digits = 3), 0.766)
  

})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

test_upsilon_by_hashing <- function() {
    invisible(.Call('_GridOnClusters_test_upsilon_by_hashing', PACKAGE = 'GridOnClusters'))
}

fast_VII_mclust <- function(X, K_range, threads = 1L, min_pts = 5L, verbose = FALSE) {
    .Call('_GridOnClusters_fast_VII_mclust', PACKAGE = 'GridOnClusters', X, K_range, threads, min_pts, verbose)
}

median_distance_c <- function(data, labels) {
    .Call('_GridOnClusters_median_distance_c', PACKAGE = 'GridOnClusters', data, labels)
}

mean_distance_c <- function(data, labels) {
    .Call('_GridOnClusters_mean_distance_c', PACKAGE = 'GridOnClusters', data, labels)
}

findgrid <- function(cluster_info, k, nobs, ndims, min_bins, max_bins, method, cutoff = 0.05, entropy = FALSE, reduction = FALSE, n_thread = 1L) {
    .Call('_GridOnClusters_findgrid', PACKAGE = 'GridOnClusters', cluster_info, k, nobs, ndims, min_bins, max_bins, method, cutoff, entropy, reduction, n_thread)
}

upsilon_c <- function(data, log) {
    .Call('_GridOnClusters_upsilon_c', PACKAGE = 'GridOnClusters', data, log)
}

---
title: "Examples of joint grid discretization"
author: "Jiandong Wang, Sajal Kumar, and Joe Song"
date: 2025-12-12
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Examples of joint grid discretization}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}

knitr::opts_chunk$set(
   collapse = TRUE,
   comment = "#>"
)
```

### History

Updated: 2025-12-12; 2022-01-27; 2022-01-17; 2020-09-13; 2020-08-01

Created: 2020-03-17


## Example 1. Nonlinear curves using kmeans+silhouette and Ball+BIC clustering with a fixed number of clusters

```{r clustering nonlinear patterns by fixed numbers of clusters, out.width="40%", fig.show="hold", fig.cap="Example 1. Nonlinear curves using kmeans+silhouette and Ball+BIC clustering with a fixed number of clusters."}
require(GridOnClusters)
x = rnorm(500)
y = sin(x)+rnorm(500, sd = 0)
z = cos(x)+rnorm(500, sd = 0)
data = cbind(x, y, z)
ks = 2:20

res = discretize.jointly(data, k=ks, cluster_method = "kmeans+silhouette", 
                         grid_method = "Sort+split") 
plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "kmeans+silhouette", 
  grid_method = "DP exact likelihood") 
plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "Ball+BIC", 
  grid_method = "DP approx likelihood 1-way")
plot(res)
```

## Example 2. Nonlinear curves and patterns using kmeans+silhouette and Ball+BIC clustering with a range for the number of clusters

```{r clustering nonlinear patterns by varying numbers of clusters, out.width="40%", fig.show="hold", fig.cap="Example 2. Using a range for the number of kmeans+silhouette and Ball+BIC clusters"}
x = rnorm(100)
y = log1p(abs(x))
z = ifelse(x >= -0.5 & x <= 0.5, 0, 1) + rnorm(100, 0, 0.1)
data = cbind(x, y, z)
ks = c(2:5)

#res = discretize.jointly(data, k=ks, cluster_method = "Ball+BIC",
#                         grid_method = "Sort+split", min_level = 1)
#plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "kmeans+silhouette", 
  grid_method = "Sort+split", min_level = 1) 

plot(res)

#res = discretize.jointly(data, k=ks, cluster_method = "kmeans+silhouette", 
#                         grid_method = "DP exact likelihood", min_level = 1) 
#plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "Ball+BIC", 
  grid_method = "DP exact likelihood", min_level = 1) 
plot(res)
```

## Example 3. Using the partition around medoids clustering method

```{r Example 3 using PAM for clustering, out.width="40%", fig.show="hold", fig.cap="Example 3. Using the partition around medoids clustering method."}
# using a clustering method other than kmeans+silhouette
x = rnorm(100)
y = log1p(abs(x))
z = sin(x)
data = cbind(x, y, z)

# pre-cluster the data using partition around medoids (PAM)
cluster_label = cluster::pam(x=data, diss = FALSE, metric = "euclidean", k = 4)$clustering

res = discretize.jointly(
  data, cluster_label = cluster_label,
  grid_method = "Sort+split", min_level = 1)
res = discretize.jointly(
  data, cluster_label = cluster_label,
  grid_method = "DP exact likelihood", min_level = 1)
plot(res, main="Original data\nPAM clustering", 
     main.table="Discretized data\nPAM & Sort+split")
```

## Example 4 Random patterns using kmeans+silhouette and Ball+BIC clustering with a range for the number of clusters

```{r Example 4, out.width="40%", fig.show="hold", fig.cap="Example 4. Random patterns using kmeans+silhouette and Ball+BIC clustering with a range."}
ks = 2:20
n = 40*10
sd = 60*4

x=rnorm(2*n, sd=sd)
y=rnorm(2*n, sd=sd)
x=c(x,rnorm(2*n, sd=sd/3))
y=c(y,rnorm(2*n, sd=sd/3)+200)

data = cbind(x, y)

res = discretize.jointly(
  data, k=ks, cluster_method = "Ball+BIC",
  grid_method = "Sort+split", min_level = 1)
plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "kmeans+silhouette", 
  grid_method = "Sort+split", min_level = 1) 
plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "kmeans+silhouette", 
  grid_method = "DP approx likelihood 1-way", min_level = 1) 
plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "Ball+BIC",
  grid_method = "DP approx likelihood 1-way", min_level = 1)
plot(res)
```

## Example 5. Multi-cluster random patterns using kmeans+silhouette and Ball+BIC clustering with a range for the number of clusters

```{r Example 5 bivariate, out.width="40%", fig.show="hold", fig.cap="Example 5. Multi-cluster random patterns using kmeans+silhouette and Ball+BIC clustering with a range."}
n <- 50*8
ks <- 2:20

X.C1 <- matrix(
   c(rnorm(n, 5, sd=2),
     rnorm(n, 0, sd=40)), 
   ncol = 2, byrow = FALSE
)
X.C2 <- matrix(
   c(rnorm(n, 70, sd=1),
     rnorm(n, 0, sd=1)), 
   ncol = 2, byrow = FALSE
)

X.C3 <- matrix(
   c(rnorm(n, 150, sd=30),
     rnorm(n, 0, sd=30)), 
   ncol = 2, byrow = FALSE
)

data = rbind(X.C1, X.C3)

res = discretize.jointly(
  data, k=ks, cluster_method = "Ball+BIC",
  grid_method = "Sort+split", min_level = 1)
plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "kmeans+silhouette", 
  grid_method = "Sort+split", min_level = 1) 
plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "kmeans+silhouette", 
  grid_method = "DP approx likelihood 1-way", 
  min_level = 1, cutoff = 1) 
plot(res)
res = discretize.jointly(
  data, k=ks, cluster_method = "Ball+BIC", 
  grid_method = "DP approx likelihood 1-way", 
  min_level = 1) 
plot(res)
```

## Example 6. Exclusive or.

```{r Example 6. Exclusive or, out.width="40%", fig.show="hold", fig.cap="Example 6. Exclusive or. Dim1 ⊕ Dim2, Dim3 and Dim4 are random"}
n <- 100
ks <- 2:10

label = c(rep(1, n), rep(2,n), rep(3,n), rep(4,n))

X1 = c(rnorm(n, 0, sd=2),
     rnorm(n, 0, sd=2),
     rnorm(n, 10, sd=2),
     rnorm(n, 10, sd=2))

X2 = c(rnorm(n, 10, sd=2),
     rnorm(n, 0, sd=2),
     rnorm(n, 10, sd=2),
     rnorm(n, 0, sd=2))

X3 = c(rnorm(4*n, 20, sd=10))

X4 = c(rnorm(4*n, 3, sd=20))

data = cbind(X1, X2, X3, X4)

#res = discretize.jointly(data, k=ks, cluster_method = "kmeans+silhouette", 
#                         grid_method = "DP approx likelihood", min_level = 1) 
#plot(res)
#
#res = discretize.jointly(data, k=ks, cluster_method = "kmeans+silhouette", 
#                         grid_method = "DP Compressed majority", min_level = 1) 
#plot(res)

#res = discretize.jointly(data, k=ks, cluster_method = "Ball+BIC", 
#                         grid_method = "DP exact likelihood", min_level = 1) 
#plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "Ball+BIC", 
  grid_method = "DP approx likelihood 1-way", 
  min_level = 1) 
plot(res)

res = discretize.jointly(
  data, k=ks, cluster_label = label,
  grid_method = "DP approx likelihood 1-way", 
  min_level = 1) 
plot(res)

#res = discretize.jointly(data, k=ks, cluster_label = label,
#                         grid_method = "Sort+split", min_level = 1) 
#plot(res)
#
#res = discretize.jointly(data, k=ks, cluster_method = "kmeans+silhouette", 
#                         grid_method = "Sort+split", min_level = 1) 
#plot(res)

```

<!-- ## Example 7. Toy dataset from paper -->

<!-- ```{r Example 7. Toy dataset, out.width="40%", fig.show="hold", fig.cap="Example 7. Toy dataset from paper"} -->
<!-- ks <- 2:10 -->

<!-- X1 = c(1,1,2,2,3,4,5,5,6,12) -->

<!-- X2 = c(7,6,7,6,7,1,2,1,1,5) -->

<!-- data = cbind(X1, X2) -->

<!-- res = discretize.jointly(data, k=ks, cluster_method = "Ball+BIC", -->
<!--                          grid_method = "DP exact likelihood", min_level = 1)  -->
<!-- plot(res) -->

<!-- ``` -->

<!-- ##Example 8. The testing dataset provided by Ruby, most of the points overlap with each other in this dataset -->

<!-- ```{r Example 8. Ruby\'s dataset, out.width="40%", fig.show="hold", fig.cap="Example 8. The testing dataset provided by Ruby, most of the points overlap with each other in this dataset."} -->
<!-- file = system.file("extdata", "TestforGridOnCluster.Rdata", package = "GridOnClusters") -->
<!-- load(file) -->
<!-- input = cbind(E1,E2) -->
<!-- res = discretize.jointly(data = input, k = c(2:9), min_level = 1,  -->
<!--                          cluster_method = "Ball+BIC",  -->
<!--                          grid_method = "DP approx likelihood") -->
<!-- plot(res) -->

<!-- res = discretize.jointly(data = input, k = c(2:9), min_level = 1,  -->
<!--                          cluster_method = "Ball+BIC", grid_method = "Sort+split") -->
<!-- plot(res) -->
<!-- ``` -->

## Example 7. Three well separated clusters

```{r Example 7. Three well separated rounds., out.width="40%", fig.show="hold", fig.cap="Example 7. Three rounds well seperated on x axis"}
n <- 20
ks <- 2:10

label = c(rep(1, n), rep(2,5*n), rep(3, 5*n))

X1 = c(rnorm(n, 0, sd=1), rnorm(5*n, 15, sd=3), rnorm(5*n, 35, sd=3))
X2 = c(rnorm(n, 0, sd=1), rnorm(5*n, 0, sd=3), rnorm(5*n, 0, sd=3))

data = cbind(X1, X2)

res = discretize.jointly(
  data, cluster_label = label,
  grid_method = "DP exact likelihood", min_level = 1) 
plot(res)
res = discretize.jointly(
  data,  cluster_label = label,
  grid_method = "DP approx likelihood 1-way", 
  min_level = 1) 
plot(res)

res.entropy = discretize.jointly(
  data,  cluster_label = label,
  grid_method = "DP exact likelihood", min_level = 1, entropy = TRUE) 
plot(res.entropy)


```

## Example 8. Four spheres with varying centers and radii

```{r Example 8. Four spheres with varying centers and radius, out.width="40%", fig.show="hold", fig.cap="Example 8. Four sphers with different centers and radius"}
n <- 200
ks <- 2:10

label = c(rep(1, n), rep(2,5*n), rep(3,n), rep(4,3*n))

X1 = c(rnorm(n, 0, sd=3), rnorm(5*n, 15, sd=3), rnorm(n, 5, sd=2), rnorm(3*n, 15, sd=4))

X2 = c(rnorm(n, 0, sd=4), rnorm(5*n, 15, sd=3), rnorm(n, 10, sd=2), rnorm(3*n, 0, sd=4))

data = cbind(X1, X2)

res = discretize.jointly(
  data, k=ks, cluster_method = "Ball+BIC",
  grid_method = "DP exact likelihood", min_level = 1) 
plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "Ball+BIC",
  grid_method = "DP approx likelihood 1-way", 
  min_level = 1) 
plot(res)

res = discretize.jointly(
  data, k=ks, cluster_method = "Ball+BIC",
  grid_method = "DP exact likelihood", min_level = 1, entropy = TRUE) 
plot(res)


```

## Example 9. A small dense sphere overlapping a large sphere

Here we evaluate whether a method can recognize the two spheres.

```{r Example 9. Two overlapping spheres, out.width="40%", fig.show="hold", fig.cap="Example 9. A small dense sphere overlapping a large sphere."}
n <- 200
ks <- 2:10

label = c(rep(1, n), rep(2,5*n))

X1 = c(rnorm(n, 0, sd=1), rnorm(5*n, 6, sd=3))
X2 = c(rnorm(n, 0, sd=1), rnorm(5*n, 0, sd=3))

data = cbind(X1, X2)

res = discretize.jointly(
  data, k=ks, cluster_label = label,
  grid_method = "DP exact likelihood", min_level = 1) 
plot(res)

res = discretize.jointly(
  data, k=ks,  cluster_label = label,
  grid_method = "DP approx likelihood 1-way", 
  min_level = 1) 
plot(res)

res.entropy = discretize.jointly(
  data, k=ks,  cluster_label = label,
  grid_method = "DP exact likelihood", min_level = 1, entropy = TRUE) 
plot(res.entropy)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ClusterVII.R
\name{ballclust}
\alias{ballclust}
\title{Fast VII Gaussian Mixture Model}
\usage{
ballclust(
  data,
  G = NULL,
  threads = parallel::detectCores() - 1,
  min_pts = 5,
  verbose = FALSE
)
}
\arguments{
\item{data}{A numeric matrix or data frame.}

\item{G}{An integer vector specifying the numbers of mixture components. 
If NULL, defaults to 1:9.}

\item{threads}{Number of threads for OpenMP parallelization.}

\item{min_pts}{an integer to specify the minimum number of points inside 
a cluster. Default: 5}

\item{verbose}{logical. if \code{TRUE}, print out additional messages. 
Default: \code{FALSE}}
}
\value{
An object of class \code{ballclust}, mimicking \code{mclust}.
}
\description{
Fast VII Gaussian Mixture Model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tools.R
\name{gen_simdata}
\alias{gen_simdata}
\title{Generate Simulated Data}
\usage{
gen_simdata(cord, sim_table, noise = 0.3, plot = FALSE)
}
\arguments{
\item{cord}{data matrix that records the index for
each cluster on each dimension}

\item{sim_table}{a matrix}

\item{noise}{a numeric value to specify noise level}

\item{plot}{a logical to turn on or off plotting}
}
\description{
Generate Simulated Data
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_patterns.R
\name{plot.GridOnClusters}
\alias{plot.GridOnClusters}
\title{Plotting the continuous data along with cluster-preserving Grid}
\usage{
\method{plot}{GridOnClusters}(
  x,
  xlab = NULL,
  ylab = NULL,
  main = NULL,
  main.table = NULL,
  col,
  line_col = "black",
  cex = 1.125,
  sub = NULL,
  pch = 19,
  plot.table = TRUE,
  ...
)
}
\arguments{
\item{x}{the result generated by discretize.jointly}

\item{xlab}{the horizontal axis label}

\item{ylab}{the vertical axis label}

\item{main}{the title of the clustering scatter plots}

\item{main.table}{the title of the discretized data plots}

\item{col}{the color of data points}

\item{line_col}{the color of grid lines}

\item{cex}{A numerical value giving the amount by 
which plotting text and symbols should be magnified 
relative to the default.}

\item{sub}{the subtitle}

\item{pch}{the symbol for points on the scatter plots}

\item{plot.table}{a logical to show the contingency 
table. Default: \code{TRUE}.}

\item{...}{additional graphical parameters}
}
\description{
Plots examples of jointly discretizing continuous data based on grids that 
preserve clusters in the original data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/discretize_jointly.R
\name{discretize.jointly}
\alias{discretize.jointly}
\title{Discretize Multivariate Continuous Data by a Cluster-Preserving Grid}
\usage{
discretize.jointly(
  data,
  k = c(2:10),
  min_level = 1,
  max_level = 100,
  cluster_method = c("Ball+BIC", "mclust_VII", "kmeans+silhouette", "PAM"),
  grid_method = c("DP approx likelihood 1-way", "DP approx likelihood 2-way",
    "DP exact likelihood", "DP Compressed majority", "DP", "Sort+split",
    "MultiChannel.WUC"),
  eval_method = c("ARI", "purity", "upsllion", "CAIR"),
  cluster_label = NULL,
  cutoff = 0,
  entropy = FALSE,
  noise = FALSE,
  dim_reduction = FALSE,
  scale = FALSE,
  variance = 0.5,
  min_pts = 5,
  nthread = 1,
  debug = FALSE
)
}
\arguments{
\item{data}{a matrix containing two or more continuous variables.
Columns are variables, rows are observations.}

\item{k}{either an integer, a vector of integers, or \code{Inf}, specifying
different ways to find clusters in \code{data}. The default is a vector
containing integers from 2 to 10. If 'k' is a single number, \code{data} will
be grouped into into exactly 'k' clusters. If 'k' is an integer vector, an
optimal 'k' is chosen from among the integers, that maximizes the average
silhouette width. If 'k' is set to \code{Inf}, an optimal k is chosen among
2 to \code{nrow(data)}. If \code{cluster_label} is specified, \code{k} is
ignored.}

\item{min_level}{integer or vector, signifying the minimum number of levels
along each dimension. If a vector of size \code{ncol(data)}, then each element
will be mapped 1:1 to each dimension in order. If an integer, then all dimensions
will have the same minimum number of levels.}

\item{max_level}{integer or vector, signifying the maximum number of levels
along each dimension. Works in the same way as \code{min_level}. Max_level will
be set to the smaller one between the number of compressed zones and itself, if
the grid_method is set to "dp compressed likelihood" or "dp compressed majority"}

\item{cluster_method}{the clustering method to be used. Ignored if cluster
labels are given.

\code{"kmeans+silhouette"} will use k-means to cluster \code{data} and the average
Silhouette score to select the number of clusters k.

\code{"Ball+BIC"}: Uses a custom, multithreaded spherical clustering method.
It builds isotropic, variable-radius clusters and uses the BIC score to
automatically select the optimal k.

\code{"mclust_VII"}: Uses the \code{mclust} package (\code{modelNames = "VII"})
to perform spherical clustering, also using BIC for k selection.

\code{"PAM"} pam is partition around medoids

See Details for more information.}

\item{grid_method}{a character string to specify the discretization method.
Default: \code{"DP approx likelihood 1-way"}. The methods
can be roughly separate into three different categories:
by cluster likelihood, by density, and by SSE (Sum of Squared Errors).

See Details for more information.}

\item{eval_method}{a character string to
specify the method to evaluate quality
of discretized data}

\item{cluster_label}{a vector of user-specified cluster labels for each observation
in \code{data}. The user is free to choose any clustering.
If unspecified, k-means clustering is used by default.}

\item{cutoff}{a floating number. Tell the quality to decide whether cut that
line or not. If the statistic, a value to evaluate the quality of a cut,is
smaller then cut_off, no cut will be performed. Only will be used when the
\code{grid_method} is set to "DP" or "DP Compressed majority"}

\item{entropy}{Boolean variable decide whether using entropy or likelihood,
default value is false.}

\item{noise}{adding jitter noise to the original data, only will be used when
the \code{cluster_method} set to "Ball+BIC". When the data are very dense,
adding noise could improve the Mclust's result.}

\item{dim_reduction}{a logical to turn on/off
dimension reduction. Default: \code{FALSE}.}

\item{scale}{a logical to specify linear
scaling of the variable in each dimension.
Default: \code{FALSE}.}

\item{variance}{a numeric value to specify
noise variance to be added to the data}

\item{min_pts}{an integer to specify the minimum number of points inside
a cluster. Default: 5}

\item{nthread}{Integer. Number of CPU threads to use.
Automatically adjusted if invalid or
exceeds available cores. Default: 1}

\item{debug}{logical. If \code{TRUE}, it enables the debug mode.
Default: \code{FALSE}}
}
\value{
A list that contains four items:
\item{\code{D}}{a matrix that contains the discretized version of the original \code{data}.
Discretized values are one(1)-based.}

\item{\code{grid}}{a list of vectors containing decision boundaries for each variable/dimension.}

\item{\code{clabels}}{a vector containing cluster labels for each observation in \code{data}.}

\item{\code{csimilarity}}{a similarity score between clusters from joint discretization
\code{D} and cluster labels \code{clabels}. The score is the adjusted Rand index.}
}
\description{
Discretize multivariate continuous data using a grid that captures the joint distribution via
preserving clusters in the original data
}
\details{
The function implements both published algorithms described in
\insertCite{Jwang2020BCB}{GridOnClusters} and new algorithms for
multivariate discretization.

The grid methods provided is roughly in three different kinds.
\itemize{
\item By Density
\itemize{
\item \code{"Sort+split"} \insertCite{Jwang2020BCB}{GridOnClusters}
will sort the cluster by cluster mean in each dimension and then
split consecutive pairs only if the sum of the error rate of each cluster is
less than or equal to 50\%. It is possible that no grid line will be added
in a certain dimension. The maximum number of lines is the number of
clusters minus one.
}
\item By SSE (Sum of Squared Errors)
\itemize{
\item \code{"MultiChannel.WUC"} will split each dimension by weighted with-in cluster
sum of squared distances by "Ckmeans.1d.dp::MultiChannel.WUC". Applied in
each projection on each dimension. The channel of each point is defined by
its multivariate cluster label.
\item \code{"DP"} will order the label by the data in each dimension and then cut the data
into max of max_level bins, then evaluating the quality of each cut to find
the best number of bins.
\item \code{"DP Compressed majority"} will order the label by the data in each dimension
and then compress the labels that neighbored by the same label to avoid trying
discretization within consecutive points of the same cluster label to greatly
reduce runtime of dynamic programming. Then cut the data into max of max_level
bins, and evaluating the quality of each cut by the majority of the data has
been spliced clearly to find the best number of bins.
}
\item By cluster likelihood
\itemize{
\item \code{"DP exact likelihood"} will order the label by the data in each dimension
and then compress the labels that neighbored by the same label to avoid trying
discretization within consecutive points of the same cluster label to greatly
reduce runtime of dynamic programming. Then cut the data into max of max_level
bins.
\item \code{"DP approx likelihood 1-way"} is a speed up version of the
\code{"DP exact likelihood"} method, but no optimal will be guaranteed.
\item \code{"DP approx likelihood 2-way"} is a bidirectional variant of the
\code{"DP approx likelihood"} method. It performs approximate dynamic
programming cuts in both the forward and backward directions and selects
the better of the two results. This approach provides additional robustness
compared to the one-directional version, but optimality is still not guaranteed.
}
}

The clsuter methods provided:

\if{html}{\out{<div class="sourceCode">}}\preformatted{- \code{"Ball+BIC"}: Employs a custom, high-performance spherical clustering 
}\if{html}{\out{</div>}}

algorithm. It models clusters as isotropic spheres (equal shape, variable volume)
and automatically selects the optimal number of clusters (k) by maximizing
the Bayesian Information Criterion (BIC). This implementation supports
multithreading for scalable execution on high-dimensional data.

\if{html}{\out{<div class="sourceCode">}}\preformatted{- \code{"mclust_VII"}: Utilizes the \code{mclust} package to fit a Gaussian 
}\if{html}{\out{</div>}}

Mixture Model with a spherical, unequal volume covariance structure
(\code{modelNames = "VII"}). It relies on the BIC score to automatically
determine the optimal number of components.
}
\note{
The default value of \code{grid_method} is changed
from "Sort+Split" (up to released package version 0.1.0.2)
to "DP approx likelihood 1-way" representing a much
improved approach.

For the default configuration, it will also add some jitter noise to the data randomly only for clustering
to avoid clustering failure by too many overlapping points in the data.

For more details, please check the documents.
}
\examples{
# using a specified k
x = rnorm(100)
y = sin(x)
z = cos(x)
data = cbind(x, y, z)
discretized_data = discretize.jointly(data, k=5)$D

# using a range of k
x = rnorm(100)
y = log1p(abs(x))
z = x^2 # tan(x)
data = cbind(x, y, z)
discretized_data = discretize.jointly(data, k=c(3:10))$D

# using k = Inf
x = c()
y = c()
mns = seq(0,1200,100)
for(i in 1:12){
  x = c(x,runif(n=20, min=mns[i], max=mns[i]+20))
  y = c(y,runif(n=20, min=mns[i], max=mns[i]+20))
}
data = cbind(x, y)
discretized_data = discretize.jointly(data, k=Inf)$D

# using an alternate clustering method to k-means
library(cluster)
x = rnorm(100)
y = log1p(abs(x))
z = sin(x)
data = cbind(x, y, z)

# pre-cluster the data using partition around medoids (PAM)
cluster_label = pam(x=data, diss = FALSE, metric = "euclidean", k = 5)$clustering
discretized_data = discretize.jointly(data, cluster_label = cluster_label)$D

#
}
\references{
\insertAllCited{}
}
\seealso{
See \link[Ckmeans.1d.dp]{Ckmeans.1d.dp} for discretizing univariate continuous data.
}
\author{
Jiandong Wang, Sajal Kumar and Mingzhou Song
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/discretize_jointly.R
\name{cluster}
\alias{cluster}
\title{Cluster Multivariate Data}
\usage{
cluster(data, k, method, noise, min_pts, nthread)
}
\arguments{
\item{data}{input continuous multivariate data}

\item{k}{the number(s) of clusters}

\item{method}{the method for clustering}

\item{noise}{adding jitter noise to the data or not}

\item{min_pts}{an integer to specify the minimum number of points inside a cluster.}

\item{nthread}{an integer to specify the number of threads}
}
\description{
The function obtains clusters from data using the given
  number of clusters, which may be a range.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_patterns.R
\name{plotGOCpatterns}
\alias{plotGOCpatterns}
\title{(OBOSOLETE) Plotting the continuous data along with cluster-preserving Grid}
\usage{
plotGOCpatterns(data, res)
}
\arguments{
\item{data}{the input continuous data matrix}

\item{res}{the result generated by discretize.jointly}
}
\description{
Plots examples of jointly discretizing continuous data based on grids that 
preserve clusters in the original data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tools.R
\name{hashed_upsilon}
\alias{hashed_upsilon}
\title{Hashed Upsilon Test of Association}
\usage{
hashed_upsilon(points, log.p = FALSE)
}
\arguments{
\item{points}{a matrix of observed categorical data points. 
Rows are observations and columns are variables.}

\item{log.p}{a logical. If \code{TRUE}, the returned p.value is 
logarithm of the true p-value. Useful for improved precision
when true p-values are close to zero.}
}
\description{
Hashed Upsilon Test of Association
}
\seealso{
\link[Upsilon]{upsilon.test} for full description of 
  the Upsilon test.
}
